# helper function for join aggregations

from typing import Hashable

import numpy as np
import pandas as pd

from janitor.functions._conditional_join import _agg_functions, _helpers


def _agg_join_left(df: pd.DataFrame, aggfunc: list, indices: dict) -> pd.DataFrame:
    """
    Compute aggregation for multiple joins
    """
    if not indices["left_index"].size:
        dtypes = df.dtypes
        aggs = {}
        for column_name, agg in aggfunc:
            if agg == "size":
                _dtype = "int64"
            else:
                _dtype = dtypes.loc[column_name]
            out = pd.array([], dtype=_dtype, copy=False)
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
        return pd.DataFrame(aggs, copy=False)
    # single join - less than
    if (indices.get("starts") is not None) and isinstance(indices.get("ends"), int):
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_rev_starts,
            "size": _agg_functions._size_rev_starts,
            "min": _agg_functions._min_rev_starts,
            "max": _agg_functions._max_rev_starts,
            "prod": _agg_functions._prod_rev_starts,
        }
        length = indices["ends"] - indices["starts"].min()
        for column_name, agg in aggfunc:
            if agg == "size":
                func = mapping[agg]
                _index, out = func(
                    starts=indices["starts"],
                    index=indices["right_index"],
                    length=length,
                )
            else:
                ser = df.loc[indices["left_index"], column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                _index, out = func(
                    arr=arr,
                    starts=indices["starts"],
                    index=indices["right_index"],
                    booleans=booleans,
                    length=length,
                )
                if agg in {
                    "sum",
                    "prod",
                } and pd.api.types.is_extension_array_dtype(ser):
                    out = pd.array(out, dtype=ser.dtype)
                elif agg in {"min", "max"}:
                    bools = out == -1
                    out = ser.iloc[out]
                    if bools.any():
                        out = out.mask(bools)
                    out = out._values
            out = pd.Series(out, index=_index)
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    # single join - greater than
    elif (indices.get("ends") is not None) and isinstance(indices.get("starts"), int):
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_rev_ends,
            "size": _agg_functions._size_rev_ends,
            "min": _agg_functions._min_rev_ends,
            "max": _agg_functions._max_rev_ends,
            "prod": _agg_functions._prod_rev_ends,
        }
        length = indices["ends"].max()
        for column_name, agg in aggfunc:
            if agg == "size":
                func = mapping[agg]
                _index, out = func(
                    ends=indices["ends"],
                    index=indices["right_index"],
                    length=length,
                )
            else:
                ser = df.loc[indices["left_index"], column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                _index, out = func(
                    arr=arr,
                    ends=indices["ends"],
                    index=indices["right_index"],
                    booleans=booleans,
                    length=length,
                )
                if agg in {
                    "sum",
                    "prod",
                } and pd.api.types.is_extension_array_dtype(ser):
                    out = pd.array(out, dtype=ser.dtype)
                elif agg in {"min", "max"}:
                    bools = out == -1
                    out = ser.iloc[out]
                    if bools.any():
                        out = out.mask(bools)
                    out = out._values
            out = pd.Series(out, index=_index)
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    # applies to equi join results
    # for many-to-one, or one-to-one
    elif not indices.keys() - {"left_index", "right_index"}:
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_rev_no_ranges,
            "min": _agg_functions._min_rev_no_ranges,
            "max": _agg_functions._max_rev_no_ranges,
            "prod": _agg_functions._prod_rev_no_ranges,
        }
        for column_name, agg in aggfunc:
            if agg == "size":
                out = pd.Series(indices["right_index"]).value_counts(sort=False)
            else:
                ser = df[column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                _index, out = func(
                    arr=arr,
                    left_index=indices["left_index"],
                    right_index=indices["right_index"],
                    booleans=booleans,
                    length=indices["right_index"].size,
                )
                if agg in {
                    "sum",
                    "prod",
                } and pd.api.types.is_extension_array_dtype(ser):
                    out = pd.array(out, dtype=ser.dtype)
                elif agg in {"min", "max"}:
                    bools = out == -1
                    out = ser.iloc[out]
                    if bools.any():
                        out = out.mask(bools)
                    out = out._values
                out = pd.Series(out, index=_index)
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    elif (indices.get("starts") is not None) and (indices.get("ends") is None):
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_rev_starts_matches,
            "size": _agg_functions._size_rev_starts_matches,
            "min": _agg_functions._min_rev_starts_matches,
            "max": _agg_functions._max_rev_starts_matches,
            "prod": _agg_functions._prod_rev_starts_matches,
        }
        for column_name, agg in aggfunc:
            if agg == "size":
                func = mapping[agg]
                _index, out = func(
                    starts=indices["starts"],
                    index=indices["right_index"],
                    matches=indices["matches"],
                    length=indices["right_index"].size,
                )
            else:
                ser = df.loc[indices["left_index"], column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                _index, out = func(
                    arr=arr,
                    starts=indices["starts"],
                    index=indices["right_index"],
                    matches=indices["matches"],
                    counts=indices["counts_array"],
                    booleans=booleans,
                    length=indices["right_index"].size,
                )
                if agg in {
                    "sum",
                    "prod",
                } and pd.api.types.is_extension_array_dtype(ser):
                    out = pd.array(out, dtype=ser.dtype)
                elif agg in {"min", "max"}:
                    bools = out == -1
                    out = ser.iloc[out]
                    if bools.any():
                        out = out.mask(bools)
                    out = out._values
            out = pd.Series(out, index=_index)
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    elif (indices.get("starts") is None) and (indices.get("ends") is not None):
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_rev_ends_matches,
            "size": _agg_functions._size_rev_ends_matches,
            "min": _agg_functions._min_rev_ends_matches,
            "max": _agg_functions._max_rev_ends_matches,
            "prod": _agg_functions._prod_rev_ends_matches,
        }
        length = indices["ends"].max()
        for column_name, agg in aggfunc:
            if agg == "size":
                func = mapping[agg]
                _index, out = func(
                    ends=indices["ends"],
                    index=indices["right_index"],
                    matches=indices["matches"],
                    length=length,
                )
            else:
                ser = df.loc[indices["left_index"], column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                _index, out = func(
                    arr=arr,
                    ends=indices["ends"],
                    index=indices["right_index"],
                    matches=indices["matches"],
                    counts=indices["counts_array"],
                    booleans=booleans,
                    length=length,
                )
                if agg in {
                    "sum",
                    "prod",
                } and pd.api.types.is_extension_array_dtype(ser):
                    out = pd.array(out, dtype=ser.dtype)
                elif agg in {"min", "max"}:
                    bools = out == -1
                    out = ser.iloc[out]
                    if bools.any():
                        out = out.mask(bools)
                    out = out._values
            out = pd.Series(out, index=_index)
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    elif indices.get("positions") is not None:
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_rev_positions,
            "size": _agg_functions._size_rev_positions,
            "min": _agg_functions._min_rev_positions,
            "max": _agg_functions._max_rev_positions,
            "prod": _agg_functions._prod_rev_positions,
        }
        for column_name, agg in aggfunc:
            if agg == "size":
                func = mapping[agg]
                _index, out = func(
                    starts=indices["starts"],
                    ends=indices["ends"],
                    positions=indices["positions"],
                    index=indices["right_index"],
                    length=indices["right_index"].size,
                )
            else:
                ser = df.loc[indices["left_index"], column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                _index, out = func(
                    arr=arr,
                    starts=indices["starts"],
                    ends=indices["ends"],
                    positions=indices["positions"],
                    index=indices["right_index"],
                    booleans=booleans,
                    length=indices["right_index"].size,
                )
                if agg in {
                    "sum",
                    "prod",
                } and pd.api.types.is_extension_array_dtype(ser):
                    out = pd.array(out, dtype=ser.dtype)
                elif agg in {"min", "max"}:
                    bools = out == -1
                    out = ser.iloc[out]
                    if bools.any():
                        out = out.mask(bools)
                    out = out._values
            out = pd.Series(out, index=_index)
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    elif (
        (indices.get("starts") is not None)
        and (indices.get("ends") is not None)
        and (indices.get("matches", None) is None)
    ):
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_rev_starts_ends,
            "size": _agg_functions._size_rev_starts_ends,
            "min": _agg_functions._min_rev_starts_ends,
            "max": _agg_functions._max_rev_starts_ends,
            "prod": _agg_functions._prod_rev_starts_ends,
        }
        length = indices["ends"].max() - indices["starts"].min()
        for column_name, agg in aggfunc:
            if agg == "size":
                func = mapping[agg]
                _index, out = func(
                    starts=indices["starts"],
                    ends=indices["ends"],
                    index=indices["right_index"],
                    length=length,
                )
            else:
                ser = df.loc[indices["left_index"], column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                _index, out = func(
                    arr=arr,
                    starts=indices["starts"],
                    ends=indices["ends"],
                    index=indices["right_index"],
                    booleans=booleans,
                    length=length,
                )
                if agg in {
                    "sum",
                    "prod",
                } and pd.api.types.is_extension_array_dtype(ser):
                    out = pd.array(out, dtype=ser.dtype)
                elif agg in {"min", "max"}:
                    bools = out == -1
                    out = ser.iloc[out]
                    if bools.any():
                        out = out.mask(bools)
                    out = out._values
            out = pd.Series(out, index=_index)
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    else:
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_rev_starts_ends_matches,
            "size": _agg_functions._size_rev_starts_ends_matches,
            "min": _agg_functions._min_rev_starts_ends_matches,
            "max": _agg_functions._max_rev_starts_ends_matches,
            "prod": _agg_functions._prod_rev_starts_ends_matches,
        }
        length = indices["ends"].max() - indices["starts"].min()
        for column_name, agg in aggfunc:
            if agg == "size":
                func = mapping[agg]
                _index, out = func(
                    starts=indices["starts"],
                    ends=indices["ends"],
                    index=indices["right_index"],
                    matches=indices["matches"],
                    length=length,
                )
            else:
                ser = df.loc[indices["left_index"], column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                _index, out = func(
                    arr=arr,
                    starts=indices["starts"],
                    ends=indices["ends"],
                    matches=indices["matches"],
                    index=indices["right_index"],
                    counts=indices["counts_array"],
                    booleans=booleans,
                    length=length,
                )
                if agg in {
                    "sum",
                    "prod",
                } and pd.api.types.is_extension_array_dtype(ser):
                    out = pd.array(out, dtype=ser.dtype)
                elif agg in {"min", "max"}:
                    bools = out == -1
                    out = ser.iloc[out]
                    if bools.any():
                        out = out.mask(bools)
                    out = out._values
            out = pd.Series(out, index=_index)
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    return pd.DataFrame(aggs, copy=False, index=_index)


def _agg_join_right(right: pd.DataFrame, aggfunc: list, indices: dict) -> pd.DataFrame:
    """
    Compute aggregation for multiple joins
    """
    if not indices["left_index"].size:
        dtypes = right.dtypes
        aggs = {}
        for column_name, agg in aggfunc:
            if agg == "size":
                _dtype = "int64"
            else:
                _dtype = dtypes.loc[column_name]
            out = pd.array([], dtype=_dtype, copy=False)
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
        return pd.DataFrame(aggs, copy=False)
    # single join - less than
    if (indices.get("starts") is not None) and isinstance(indices.get("ends"), int):
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_starts,
            "min": _agg_functions._min_starts,
            "max": _agg_functions._max_starts,
            "prod": _agg_functions._prod_starts,
        }
        for column_name, agg in aggfunc:
            if agg == "size":
                out = indices["ends"] - indices["starts"]
            else:
                ser = right.loc[indices["right_index"], column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                out = func(arr=arr, starts=indices["starts"], booleans=booleans)
                if agg in {
                    "sum",
                    "prod",
                } and pd.api.types.is_extension_array_dtype(ser):
                    out = pd.array(out, dtype=ser.dtype)
                elif agg in {"min", "max"}:
                    bools = out == -1
                    out = ser.iloc[out]
                    if bools.any():
                        out = out.mask(bools)
                    out = out._values
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    # single join - greater than
    elif (indices.get("ends") is not None) and isinstance(indices.get("starts"), int):
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_ends,
            "min": _agg_functions._min_ends,
            "max": _agg_functions._max_ends,
            "prod": _agg_functions._prod_ends,
        }
        for column_name, agg in aggfunc:
            if agg == "size":
                out = indices["ends"]
            else:
                ser = right.loc[indices["right_index"], column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                out = func(arr=arr, ends=indices["ends"], booleans=booleans)
                if agg in {
                    "sum",
                    "prod",
                } and pd.api.types.is_extension_array_dtype(ser):
                    out = pd.array(out, dtype=ser.dtype)
                elif agg in {"min", "max"}:
                    bools = out == -1
                    out = ser.iloc[out]
                    if bools.any():
                        out = out.mask(bools)
                    out = out._values
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    # applies to equi join results
    # for many-to-one, or one-to-one
    elif not indices.keys() - {"left_index", "right_index"}:
        aggs = {}
        right_index = indices["right_index"]
        for column_name, agg in aggfunc:
            if agg == "size":
                out = np.repeat(1, right_index.size)
            else:
                out = right.loc[right_index, column_name]
                out = out._values
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    elif (indices.get("starts") is not None) and (indices.get("ends") is None):
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_starts_matches,
            "min": _agg_functions._min_starts_matches,
            "max": _agg_functions._max_starts_matches,
            "prod": _agg_functions._prod_starts_matches,
        }
        for column_name, agg in aggfunc:
            if agg == "size":
                out = indices["counts_array"]
            else:
                ser = right.loc[indices["right_index"], column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                out = func(
                    arr=arr,
                    starts=indices["starts"],
                    counts=indices["counts_array"],
                    matches=indices["matches"],
                    booleans=booleans,
                )
            if agg in {
                "sum",
                "prod",
            } and pd.api.types.is_extension_array_dtype(ser):
                out = pd.array(out, dtype=ser.dtype)
            elif agg in {"min", "max"}:
                bools = out == -1
                out = ser.iloc[out]
                if bools.any():
                    out = out.mask(bools)
                out = out._values
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    elif (indices.get("starts") is None) and (indices.get("ends") is not None):
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_ends_matches,
            "min": _agg_functions._min_ends_matches,
            "max": _agg_functions._max_ends_matches,
            "prod": _agg_functions._prod_ends_matches,
        }
        for column_name, agg in aggfunc:
            if agg == "size":
                out = indices["counts_array"]
            else:
                ser = right.loc[indices["right_index"], column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                out = func(
                    arr=arr,
                    ends=indices["ends"],
                    counts=indices["counts_array"],
                    matches=indices["matches"],
                    booleans=booleans,
                )
            if agg in {
                "sum",
                "prod",
            } and pd.api.types.is_extension_array_dtype(ser):
                out = pd.array(out, dtype=ser.dtype)
            elif agg in {"min", "max"}:
                bools = out == -1
                out = ser.iloc[out]
                if bools.any():
                    out = out.mask(bools)
                out = out._values
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    elif indices.get("positions") is not None:
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_positions,
            "min": _agg_functions._min_positions,
            "max": _agg_functions._max_positions,
            "prod": _agg_functions._prod_positions,
        }
        starts = indices["starts"]
        ends = indices["ends"]
        positions = indices["positions"]
        counts_array = indices["counts_array"]
        right_index = indices["right_index"]
        for column_name, agg in aggfunc:
            if agg == "size":
                out = counts_array
            else:
                ser = right.loc[right_index, column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                out = func(
                    arr=arr,
                    starts=starts,
                    ends=ends,
                    positions=positions,
                    booleans=booleans,
                )
            if agg in {
                "sum",
                "prod",
            } and pd.api.types.is_extension_array_dtype(ser):
                out = pd.array(out, dtype=ser.dtype)
            elif agg in {"min", "max"}:
                bools = out == -1
                out = ser.iloc[out]
                if bools.any():
                    out = out.mask(bools)
                out = out._values
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    elif (
        (indices.get("starts") is not None)
        and (indices.get("ends") is not None)
        and (indices.get("matches", None) is None)
    ):
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_starts_ends,
            "min": _agg_functions._min_starts_ends,
            "max": _agg_functions._max_starts_ends,
            "prod": _agg_functions._prod_starts_ends,
        }
        starts = indices["starts"]
        ends = indices["ends"]
        right_index = indices["right_index"]
        for column_name, agg in aggfunc:
            if agg == "size":
                out = ends - starts
            else:
                ser = right.loc[right_index, column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                out = func(
                    arr=arr,
                    starts=starts,
                    ends=ends,
                    booleans=booleans,
                )
            if agg in {
                "sum",
                "prod",
            } and pd.api.types.is_extension_array_dtype(ser):
                out = pd.array(out, dtype=ser.dtype)
            elif agg in {"min", "max"}:
                bools = out == -1
                out = ser.iloc[out]
                if bools.any():
                    out = out.mask(bools)
                out = out._values
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    else:
        aggs = {}
        mapping = {
            "sum": _agg_functions._sum_starts_ends_matches,
            "min": _agg_functions._min_starts_ends_matches,
            "max": _agg_functions._max_starts_ends_matches,
            "prod": _agg_functions._prod_starts_ends_matches,
        }
        starts = indices["starts"]
        ends = indices["ends"]
        counts_array = indices["counts_array"]
        matches = indices["matches"]
        right_index = indices["right_index"]
        for column_name, agg in aggfunc:
            if agg == "size":
                out = counts_array
            else:
                ser = right.loc[right_index, column_name]
                arr = ser._values
                booleans = pd.isna(arr)
                arr = _helpers._convert_array_to_numpy(array=arr)
                func = mapping[agg]
                out = func(
                    arr=arr,
                    starts=starts,
                    ends=ends,
                    matches=matches,
                    counts=counts_array,
                    booleans=booleans,
                )
            if agg in {
                "sum",
                "prod",
            } and pd.api.types.is_extension_array_dtype(ser):
                out = pd.array(out, dtype=ser.dtype)
            elif agg in {"min", "max"}:
                bools = out == -1
                out = ser.iloc[out]
                if bools.any():
                    out = out.mask(bools)
                out = out._values
            new_label = _build_agg_label(column_name=column_name, agg_name=agg)
            aggs[new_label] = out
    return pd.DataFrame(aggs, copy=False, index=indices["left_index"])


def _build_agg_label(column_name: Hashable, agg_name: str):
    if isinstance(column_name, tuple):
        return (*column_name, agg_name)
    return (f"{column_name}", agg_name)



"""Implementation of limit_column_characters."""

import pandas as pd
import pandas_flavor as pf

from janitor.utils import check


@pf.register_dataframe_method
def limit_column_characters(
    df: pd.DataFrame,
    column_length: int,
    col_separator: str = "_",
) -> pd.DataFrame:
    """Truncate column sizes to a specific length.

    This method mutates the original DataFrame.

    Method chaining will truncate all columns to a given length and append
    a given separator character with the index of duplicate columns, except
    for the first distinct column name.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> data_dict = {
        ...     "really_long_name": [9, 8, 7],
        ...     "another_really_long_name": [2, 4, 6],
        ...     "another_really_longer_name": list("xyz"),
        ...     "this_is_getting_out_of_hand": list("pqr"),
        ... }
        >>> df = pd.DataFrame(data_dict)
        >>> df  # doctest: +SKIP
           really_long_name  another_really_long_name another_really_longer_name this_is_getting_out_of_hand
        0                 9                         2                          x                           p
        1                 8                         4                          y                           q
        2                 7                         6                          z                           r
        >>> df.limit_column_characters(7)
           really_  another another_1 this_is
        0        9        2         x       p
        1        8        4         y       q
        2        7        6         z       r

    Args:
        df: A pandas DataFrame.
        column_length: Character length for which to truncate all columns.
            The column separator value and number for duplicate column name does
            not contribute. Therefore, if all columns are truncated to 10
            characters, the first distinct column will be 10 characters and the
            remaining will be 12 characters (assuming a column separator of one
            character).
        col_separator: The separator to use for counting distinct column
            values, for example, `'_'` or `'.'`.
            Supply an empty string (i.e. `''`) to remove the separator.

    Returns:
        A pandas DataFrame with truncated column lengths.
    """  # noqa: E501

    check("column_length", column_length, [int])
    check("col_separator", col_separator, [str])

    col_names = df.columns
    col_names = [col_name[:column_length] for col_name in col_names]

    col_name_set = set(col_names)
    col_name_count = {}

    # If no columns are duplicates, we can skip the loops below.
    if len(col_name_set) == len(col_names):
        df.columns = col_names
        return df

    for col_name_to_check in col_name_set:
        count = 0
        for idx, col_name in enumerate(col_names):
            if col_name_to_check == col_name:
                col_name_count[idx] = count
                count += 1

    final_col_names = []
    for idx, col_name in enumerate(col_names):
        if col_name_count[idx] > 0:
            col_name_to_append = col_name + col_separator + str(col_name_count[idx])
            final_col_names.append(col_name_to_append)
        else:
            final_col_names.append(col_name)

    df.columns = final_col_names
    return df



"""Implementation source for `expand_grid`."""

from __future__ import annotations

from collections import defaultdict
from functools import singledispatch
from typing import Optional, Union

import numpy as np
import pandas as pd
import pandas_flavor as pf
from pandas.api.types import is_scalar
from pandas.core.common import apply_if_callable
from pandas.core.dtypes.concat import concat_compat

from janitor.functions.utils import _computations_expand_grid
from janitor.utils import check, check_column, refactored_function


@pf.register_dataframe_method
@refactored_function(
    message=(
        "This function will be deprecated in a 1.x release. "
        "Please use `janitor.cartesian_product` instead."
    )
)
def expand_grid(
    df: Optional[pd.DataFrame] = None,
    df_key: Optional[str] = None,
    *,
    others: Optional[dict] = None,
) -> Union[pd.DataFrame, None]:
    """
    Creates a DataFrame from a cartesian combination of all inputs.

    !!!note

        This function will be deprecated in a 1.x release;
        use [`cartesian_product`][janitor.functions.expand_grid.cartesian_product]
        instead.

    It is not restricted to a pandas DataFrame;
    it can work with any list-like structure
    that is 1 or 2 dimensional.

    If method-chaining to a DataFrame, a string argument
    to `df_key` parameter must be provided.

    Data types are preserved in this function,
    including pandas' extension array dtypes.

    The output will always be a DataFrame, usually with a MultiIndex column,
    with the keys of the `others` dictionary serving as the top level columns.

    If a pandas Series/DataFrame is passed, and has a labeled index, or
    a MultiIndex index, the index is discarded; the final DataFrame
    will have a RangeIndex.

    The MultiIndexed DataFrame can be flattened using pyjanitor's
    [`collapse_levels`][janitor.functions.collapse_levels.collapse_levels]
    method; the user can also decide to drop any of the levels, via pandas'
    `droplevel` method.

    Examples:
        >>> import pandas as pd
        >>> import janitor as jn
        >>> df = pd.DataFrame({"x": [1, 2], "y": [2, 1]})
        >>> data = {"z": [1, 2, 3]}
        >>> df.expand_grid(df_key="df", others=data)
          df     z
           x  y  0
        0  1  2  1
        1  1  2  2
        2  1  2  3
        3  2  1  1
        4  2  1  2
        5  2  1  3

        `expand_grid` works with non-pandas objects:

        >>> data = {"x": [1, 2, 3], "y": [1, 2]}
        >>> jn.expand_grid(others=data)
           x  y
           0  0
        0  1  1
        1  1  2
        2  2  1
        3  2  2
        4  3  1
        5  3  2

    Args:
        df: A pandas DataFrame.
        df_key: Name of key for the dataframe.
            It becomes part of the column names of the dataframe.
        others: A dictionary that contains the data
            to be combined with the dataframe.
            If no dataframe exists, all inputs
            in `others` will be combined to create a DataFrame.

    Raises:
        KeyError: If there is a DataFrame and `df_key` is not provided.

    Returns:
        A pandas DataFrame of the cartesian product.
        If `df` is not provided, and `others` is not provided,
        None is returned.
    """  # noqa: E501

    if df is not None:
        check("df", df, [pd.DataFrame])
        if not df_key:
            raise KeyError(
                "Using `expand_grid` as part of a "
                "DataFrame method chain requires that "
                "a string argument be provided for "
                "the `df_key` parameter. "
            )

        check("df_key", df_key, [str])

    if not others and (df is not None):
        return df

    if not others:
        return None

    check("others", others, [dict])

    for key in others:
        check("key", key, [str])

    if df is not None:
        others = {**{df_key: df}, **others}

    others = _computations_expand_grid(others)
    return pd.DataFrame(others, copy=False)


@pf.register_dataframe_method
def expand(
    df: pd.DataFrame,
    *columns: tuple,
    sort: bool = False,
    by: str | list = None,
) -> pd.DataFrame:
    """
    Creates a DataFrame from a cartesian combination of all inputs.

    Inspiration is from tidyr's expand() function.

    expand() is often useful with
    [pd.merge](https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.merge.html)
    to convert implicit
    missing values to explicit missing values - similar to
    [`complete`][janitor.functions.complete.complete].

    It can also be used to figure out which combinations are missing
    (e.g identify gaps in your DataFrame).

    The variable `columns` parameter can be a column name,
    a list of column names, a pandas Index/Series/DataFrame,
    or a callable, which when applied to the DataFrame,
    evaluates to a pandas Index/Series/DataFrame.

    A dictionary can also be passed
    to the variable `columns` parameter -
    the values of the dictionary should be
    either be a 1D array
    or a callable that evaluates to a
    1D array. The array should be unique;
    no check is done to verify this.

    If `by` is present, the DataFrame is *expanded* per group.
    `by` should be a column name, or a list of column names.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> data = [
        ...     {"type": "apple", "year": 2010, "size": "XS"},
        ...     {"type": "orange", "year": 2010, "size": "S"},
        ...     {"type": "apple", "year": 2012, "size": "M"},
        ...     {"type": "orange", "year": 2010, "size": "S"},
        ...     {"type": "orange", "year": 2011, "size": "S"},
        ...     {"type": "orange", "year": 2012, "size": "M"},
        ... ]
        >>> df = pd.DataFrame(data)
        >>> df
             type  year size
        0   apple  2010   XS
        1  orange  2010    S
        2   apple  2012    M
        3  orange  2010    S
        4  orange  2011    S
        5  orange  2012    M

        Get unique observations:
        >>> df.expand("type")
             type
        0   apple
        1  orange
        >>> df.expand("size")
          size
        0   XS
        1    S
        2    M
        >>> df.expand("type", "size")
             type size
        0   apple   XS
        1   apple    S
        2   apple    M
        3  orange   XS
        4  orange    S
        5  orange    M
        >>> df.expand("type", "size", "year")
              type size  year
        0    apple   XS  2010
        1    apple   XS  2012
        2    apple   XS  2011
        3    apple    S  2010
        4    apple    S  2012
        5    apple    S  2011
        6    apple    M  2010
        7    apple    M  2012
        8    apple    M  2011
        9   orange   XS  2010
        10  orange   XS  2012
        11  orange   XS  2011
        12  orange    S  2010
        13  orange    S  2012
        14  orange    S  2011
        15  orange    M  2010
        16  orange    M  2012
        17  orange    M  2011

        Get observations that only occur in the data:
        >>> df.expand(["type", "size"])
             type size
        0   apple   XS
        1  orange    S
        2   apple    M
        3  orange    M
        >>> df.expand(["type", "size", "year"])
             type size  year
        0   apple   XS  2010
        1  orange    S  2010
        2   apple    M  2012
        3  orange    S  2011
        4  orange    M  2012

        Expand the DataFrame to include new observations:
        >>> df.expand("type", "size", {"new_year": range(2010, 2014)})
              type size  new_year
        0    apple   XS      2010
        1    apple   XS      2011
        2    apple   XS      2012
        3    apple   XS      2013
        4    apple    S      2010
        5    apple    S      2011
        6    apple    S      2012
        7    apple    S      2013
        8    apple    M      2010
        9    apple    M      2011
        10   apple    M      2012
        11   apple    M      2013
        12  orange   XS      2010
        13  orange   XS      2011
        14  orange   XS      2012
        15  orange   XS      2013
        16  orange    S      2010
        17  orange    S      2011
        18  orange    S      2012
        19  orange    S      2013
        20  orange    M      2010
        21  orange    M      2011
        22  orange    M      2012
        23  orange    M      2013

        Filter for missing observations:
        >>> combo = df.expand("type", "size", "year")
        >>> anti_join = df.merge(combo, how="right", indicator=True)
        >>> anti_join.query("_merge=='right_only'").drop(columns="_merge")
              type  year size
        1    apple  2012   XS
        2    apple  2011   XS
        3    apple  2010    S
        4    apple  2012    S
        5    apple  2011    S
        6    apple  2010    M
        8    apple  2011    M
        9   orange  2010   XS
        10  orange  2012   XS
        11  orange  2011   XS
        14  orange  2012    S
        16  orange  2010    M
        18  orange  2011    M

        Expand within each group, using `by`:
        >>> df.expand("year", "size", by="type")  # doctest: +NORMALIZE_WHITESPACE
                year size
        type
        apple   2010   XS
        apple   2010    M
        apple   2012   XS
        apple   2012    M
        orange  2010    S
        orange  2010    M
        orange  2011    S
        orange  2011    M
        orange  2012    S
        orange  2012    M

    Args:
        df: A pandas DataFrame.
        columns: Specification of columns to expand.
            It could be column labels,
             a list/tuple of column labels,
             or a pandas Index/Series/DataFrame.

            It can also be a callable;
            the callable will be applied to the
            entire DataFrame. The callable should
            return a pandas Series/Index/DataFrame.

            It can also be a dictionary,
            where the values are either a 1D array
            or a callable that evaluates to a
            1D array.
            The array should be unique;
            no check is done to verify this.
        sort: If True, sort the DataFrame.
        by: Label or list of labels to group by.

    Returns:
        A pandas DataFrame.
    """  # noqa: E501
    if by is None:
        contents = _build_pandas_objects_for_expand(df=df, columns=columns)
        return cartesian_product(*contents, sort=sort)
    if not is_scalar(by) and not isinstance(by, list):
        raise TypeError(
            "The argument to the by parameter "
            "should be a scalar or a list; "
            f"instead got {type(by).__name__}"
        )
    check_column(df, column_names=by, present=True)
    grouped = df.groupby(by=by, sort=False, dropna=False, observed=True)
    index = grouped._grouper.result_index
    dictionary = defaultdict(list)
    lengths = []
    for _, frame in grouped:
        objects = _build_pandas_objects_for_expand(df=frame, columns=columns)
        objects = _compute_cartesian_product(inputs=objects, sort=False)
        length = objects[next(iter(objects))].size
        lengths.append(length)
        for k, v in objects.items():
            dictionary[k].append(v)
    dictionary = {key: concat_compat(value) for key, value in dictionary.items()}
    index = index.repeat(lengths)
    out = pd.DataFrame(data=dictionary, index=index, copy=False)
    if sort:
        headers = out.columns.tolist()
        return out.sort_values(headers)
    return out


def _build_pandas_objects_for_expand(df: pd.DataFrame, columns: tuple) -> list:
    """
    Build pandas_objects for expand().
    These will be passed to _cartesian_product
    """
    contents = []
    for position, column in enumerate(columns):
        if is_scalar(column) or isinstance(column, tuple):
            arr = df[column].drop_duplicates()
            contents.append(arr)
        elif isinstance(column, list):
            arr = df.loc[:, column].drop_duplicates()
            contents.append(arr)
        elif isinstance(column, dict):
            for label, arr in column.items():
                arr = apply_if_callable(maybe_callable=arr, obj=df)
                arr = pd.Series(arr, name=label)
                contents.append(arr)
        elif isinstance(column, (pd.Series, pd.Index, pd.DataFrame)):
            contents.append(column)
        elif callable(column):
            arr = apply_if_callable(maybe_callable=column, obj=df)
            contents.append(arr)
        else:
            raise TypeError(
                "The arguments to the variable columns parameter "
                "should either be a column name, a list of column names, "
                "a pandas Index/Series/DataFrame, "
                "a callable that evaluates to a "
                "pandas Index/Series/DataFrame, "
                "or a dictionary, "
                "where the value is a 1D array; "
                f"instead got type {type(column).__name__} "
                f"at position {position}"
            )
    return contents


def cartesian_product(*inputs: tuple, sort: bool = False) -> pd.DataFrame:
    """Creates a DataFrame from a cartesian combination of all inputs.

    Inspiration is from tidyr's expand_grid() function.

    The input argument should be a pandas Index/Series/DataFrame,
    or a dictionary - the values of the dictionary should be
    a 1D array.

    Examples:
        >>> import pandas as pd
        >>> import janitor as jn
        >>> df = pd.DataFrame({"x": [1, 2], "y": [2, 1]})
        >>> data = pd.Series([1, 2, 3], name="z")
        >>> jn.cartesian_product(df, data)
           x  y  z
        0  1  2  1
        1  1  2  2
        2  1  2  3
        3  2  1  1
        4  2  1  2
        5  2  1  3

        `cartesian_product` also works with non-pandas objects:

        >>> data = {"x": [1, 2, 3], "y": [1, 2]}
        >>> cartesian_product(data)
           x  y
        0  1  1
        1  1  2
        2  2  1
        3  2  2
        4  3  1
        5  3  2

    Args:
        *inputs: Variable arguments. The arguments should be
            a pandas Index/Series/DataFrame, or a dictionary,
            where the values in the dictionary is a 1D array.
        sort: If True, sort the output DataFrame.

    Returns:
        A pandas DataFrame.
    """
    contents = []
    for entry in inputs:
        if isinstance(entry, dict):
            for label, value in entry.items():
                arr = pd.Series(value, name=label)
                contents.append(arr)
        else:
            contents.append(entry)
    outcome = _compute_cartesian_product(inputs=contents, sort=sort)
    # the values in the outcome dictionary are copies,
    # based on numpy indexing semantics;
    # as such, it is safe to pass copy=False
    return pd.DataFrame(data=outcome, copy=False)


def _compute_cartesian_product(inputs: tuple, sort: bool) -> dict:
    """
    Compute the cartesian product of pandas objects.
    """
    unique_names = set()
    for position, entry in enumerate(inputs):
        unique_names = _validate_pandas_object(
            entry,
            position=position,
            unique_names=unique_names,
        )

    length_of_objects = [len(pandas_object) for pandas_object in inputs]
    cartesian_lengths = np.indices(length_of_objects)
    cartesian_lengths = cartesian_lengths.reshape((len(inputs), -1))
    zipped = zip(inputs, cartesian_lengths)
    contents = {}
    for pandas_object, indexer in zipped:
        if isinstance(pandas_object, pd.DataFrame):
            if sort:
                headers = pandas_object.columns.tolist()
                pandas_object = pandas_object.sort_values(headers)
            for label, array in pandas_object.items():
                contents[label] = array._values[indexer]
        elif isinstance(pandas_object, pd.MultiIndex):
            if sort:
                pandas_object, _ = pandas_object.sortlevel()
            for label in pandas_object.names:
                array = pandas_object.get_level_values(label)._values[indexer]
                contents[label] = array
        else:
            if sort:
                pandas_object = pandas_object.sort_values()
            array = pandas_object._values[indexer]
            contents[pandas_object.name] = array

    if all(map(is_scalar, contents)):
        return contents

    lengths = (len(key) for key in contents if isinstance(key, tuple))
    lengths = max(lengths)
    others = {}
    # manage differing tuple lengths
    # or a mix of tuples and scalars
    for key, value in contents.items():
        if is_scalar(key):
            key = (key, *([""] * (lengths - 1)))
            others[key] = value
        elif len(key) == lengths:
            others[key] = value
        else:
            key = (*key, *([""] * (lengths - len(key))))
            others[key] = value
    return others


@singledispatch
def _validate_pandas_object(pandas_object, position, unique_names) -> tuple:
    """
    Validate pandas object, and ensure the names of the pandas_object
    are not duplicated.

    Args:
        pandas_object: object to be validated.
        position: position of pandas_object in sequence.
        unique_names: python set of names.

    Raises:
        TypeError: If object is not a pandas Index/Series/DataFrame.

    Returns:
        A tuple (pandas_object, unique_names, length_of_objects)
    """
    raise TypeError(
        "input should be either a Pandas DataFrame, "
        "a pandas Series, or a pandas Index; "
        f"instead the object at position {position} "
        f"is of type {type(pandas_object).__name__}"
    )


@_validate_pandas_object.register(pd.DataFrame)
def _validate_pandas_object_(  # noqa: F811
    pandas_object, position, unique_names
) -> tuple:
    """
    Validate a pandas DataFrame
    """
    for label in pandas_object:
        if label in unique_names:
            raise ValueError(
                f"Label {label} in the DataFrame at position {position} is duplicated."
            )
        unique_names.add(label)
    return unique_names


@_validate_pandas_object.register(pd.MultiIndex)
def _validate_pandas_object_(  # noqa: F811
    pandas_object, position, unique_names
) -> tuple:
    """
    Validate a pandas MultiIndex
    """
    labels = pandas_object.names
    if None in labels:
        raise ValueError(
            f"Kindly ensure all levels in the MultiIndex "
            f"at position {position} is labeled."
        )
    for label in labels:
        if label in unique_names:
            raise ValueError(
                f"Label {label} in the MultiIndex at position {position} is duplicated."
            )
        unique_names.add(label)
    return unique_names


@_validate_pandas_object.register(pd.Index)
@_validate_pandas_object.register(pd.Series)
def _validate_pandas_object_(  # noqa: F811
    pandas_object, position, unique_names
) -> tuple:
    """
    Validate a pandas MultiIndex
    """
    label = pandas_object.name
    if not label:
        raise ValueError(
            f"Kindly ensure the {type(pandas_object).__name__} "
            f"at position {position} has a name."
        )
    if label in unique_names:
        raise ValueError(
            f"Label {label} in the {type(pandas_object).__name__} "
            f"at position {position} is duplicated."
        )
    unique_names.add(label)
    return unique_names



"""
Functions to augment XArray DataArrays and Datasets with additional
functionality.
"""

from typing import Union

import numpy as np
import pandas_flavor as pf
import xarray as xr


@pf.register_xarray_dataarray_method
def clone_using(
    da: xr.DataArray,
    np_arr: np.array,
    use_coords: bool = True,
    use_attrs: bool = False,
    new_name: str = None,
) -> xr.DataArray:
    """
    Given a NumPy array, return an XArray `DataArray` which contains the same
    dimension names and (optionally) coordinates and other properties as the
    supplied `DataArray`.

    This is similar to `xr.DataArray.copy()` with more specificity for
    the type of cloning you would like to perform - the different properties
    that you desire to mirror in the new `DataArray`.

    If the coordinates from the source `DataArray` are not desired, the shape
    of the source and new NumPy arrays don't need to match.
    The number of dimensions do, however.

    Examples:
        Making a new `DataArray` from a previous one, keeping the
        dimension names but dropping the coordinates (the input NumPy array
        is of a different size):

        >>> import xarray as xr
        >>> import janitor.xarray
        >>> da = xr.DataArray(
        ...     np.zeros((512, 1024)),
        ...     dims=["ax_1", "ax_2"],
        ...     coords=dict(ax_1=np.linspace(0, 1, 512), ax_2=np.logspace(-2, 2, 1024)),
        ...     name="original",
        ... )
        >>> new_da = da.clone_using(
        ...     np.ones((4, 6)),
        ...     new_name="new_and_improved",
        ...     use_coords=False,
        ... )
        >>> new_da
        <xarray.DataArray 'new_and_improved' (ax_1: 4, ax_2: 6)> Size: 192B
        array([[1., 1., 1., 1., 1., 1.],
               [1., 1., 1., 1., 1., 1.],
               [1., 1., 1., 1., 1., 1.],
               [1., 1., 1., 1., 1., 1.]])
        Dimensions without coordinates: ax_1, ax_2

    Args:
        da: The `DataArray` supplied by the method itself.
        np_arr: The NumPy array which will be wrapped in a new `DataArray`
            given the properties copied over from the source `DataArray`.
        use_coords: If `True`, use the coordinates of the source
            `DataArray` for the coordinates of the newly-generated array.
            Shapes must match in this case. If `False`, only the number of
            dimensions must match.
        use_attrs: If `True`, copy over the `attrs` from the source
            `DataArray`.
            The data inside `attrs` itself is not copied, only the mapping.
            Otherwise, use the supplied attrs.
        new_name: If set, use as the new name of the returned `DataArray`.
            Otherwise, use the name of `da`.

    Raises:
        ValueError: If number of dimensions in `NumPy` array and
            `DataArray` do not match.
        ValueError: If shape of `NumPy` array and `DataArray`
            do not match.

    Returns:
        A `DataArray` styled like the input `DataArray` containing the
            NumPy array data.
    """

    if np_arr.ndim != da.ndim:
        raise ValueError(
            "Number of dims in the NumPy array and the DataArray must match."
        )

    if use_coords and not all(
        np_ax_len == da_ax_len for np_ax_len, da_ax_len in zip(np_arr.shape, da.shape)
    ):
        raise ValueError(
            "Input NumPy array and DataArray must have the same "
            "shape if copying over coordinates."
        )

    return xr.DataArray(
        np_arr,
        dims=da.dims,
        coords=da.coords if use_coords else None,
        attrs=da.attrs.copy() if use_attrs else None,
        name=new_name if new_name is not None else da.name,
    )


@pf.register_xarray_dataset_method
@pf.register_xarray_dataarray_method
def convert_datetime_to_number(
    da_or_ds: Union[xr.DataArray, xr.Dataset],
    time_units: str,
    dim: str = "time",
) -> Union[xr.DataArray, xr.Dataset]:
    """Convert the coordinates of a datetime axis to a human-readable float
    representation.

    Examples:
        Convert a `DataArray`'s time dimension coordinates from
        minutes to seconds:

        >>> import numpy as np
        >>> import xarray as xr
        >>> import janitor.xarray
        >>> timepoints = 5
        >>> da = xr.DataArray(
        ...     np.array([2, 8, 0, 1, 7, 7]),
        ...     dims="time",
        ...     coords=dict(time=np.arange(6) * np.timedelta64(1, "m")),
        ... )
        >>> da_minutes = da.convert_datetime_to_number("s", dim="time")
        >>> da_minutes
        <xarray.DataArray (time: 6)> Size: 48B
        array([2, 8, 0, 1, 7, 7])
        Coordinates:
          * time     (time) float64 48B 0.0 60.0 120.0 180.0 240.0 300.0

    Args:
        da_or_ds: XArray object.
        time_units: Numpy timedelta string specification for the unit you
            would like to convert the coordinates to.
        dim: The time dimension whose coordinates are datetime objects.

    Returns:
        The original XArray object with the time dimension reassigned.
    """

    times = da_or_ds.coords[dim].data / np.timedelta64(1, time_units)

    return da_or_ds.assign_coords({dim: times})



"""Implementation of move."""

from typing import Any

import numpy as np
import pandas as pd
import pandas_flavor as pf

from janitor.functions.select import _index_converter, _select_index


@pf.register_dataframe_method
def move(
    df: pd.DataFrame,
    source: Any,
    target: Any = None,
    position: str = "before",
    axis: int = 0,
) -> pd.DataFrame:
    """Changes rows or columns positions in the dataframe.

    It uses the
    [`select`][janitor.functions.select.select] syntax,
    making it easy to move blocks of rows or columns at once.

    This operation does not reset the index of the dataframe. User must
    explicitly do so.

    The dataframe must have unique column names or indices.

    Examples:
        Move a row:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": [2, 4, 6, 8], "b": list("wxyz")})
        >>> df
           a  b
        0  2  w
        1  4  x
        2  6  y
        3  8  z
        >>> df.move(source=0, target=3, position="before", axis=0)
           a  b
        1  4  x
        2  6  y
        0  2  w
        3  8  z

        Move a column:
        >>> import pandas as pd
        >>> import janitor
        >>> data = [{"a": 1, "b": 1, "c": 1, "d": "a", "e": "a", "f": "a"}]
        >>> df = pd.DataFrame(data)
        >>> df
           a  b  c  d  e  f
        0  1  1  1  a  a  a
        >>> df.move(source="a", target="c", position="after", axis=1)
           b  c  a  d  e  f
        0  1  1  1  a  a  a
        >>> df.move(source="f", target="b", position="before", axis=1)
           a  f  b  c  d  e
        0  1  a  1  1  a  a
        >>> df.move(source="a", target=None, position="after", axis=1)
           b  c  d  e  f  a
        0  1  1  a  a  a  1

        Move columns:
        >>> from pandas.api.types import is_numeric_dtype, is_string_dtype
        >>> df.move(source=is_string_dtype, target=None, position="before", axis=1)
           d  e  f  a  b  c
        0  a  a  a  1  1  1
        >>> df.move(source=is_numeric_dtype, target=None, position="after", axis=1)
           d  e  f  a  b  c
        0  a  a  a  1  1  1
        >>> df.move(
        ...     source=["d", "f"], target=is_numeric_dtype, position="before", axis=1
        ... )
           d  f  a  b  c  e
        0  a  a  1  1  1  a

    Args:
        df: The pandas DataFrame object.
        source: Columns or rows to move.
        target: Columns or rows to move adjacent to.
            If `None` and `position == 'before'`, `source`
            is moved to the beginning; if `position == 'after'`,
            `source` is moved to the end.
        position: Specifies the destination of the columns/rows.
            Values can be either `before` or `after`; defaults to `before`.
        axis: Axis along which the function is applied. 0 to move along
            the index, 1 to move along the columns.

    Raises:
        ValueError: If `axis` is not `0` or `1`.
        ValueError: If `position` is not `before` or `after`.

    Returns:
        The dataframe with the Series moved.
    """  # noqa: E501
    if axis not in [0, 1]:
        raise ValueError(f"Invalid axis '{axis}'. Can only be 0 or 1.")

    if position not in ["before", "after"]:
        raise ValueError(
            f"Invalid position '{position}'. Can only be 'before' or 'after'."
        )

    mapping = {0: "index", 1: "columns"}
    names = getattr(df, mapping[axis])

    assert names.is_unique

    index = np.arange(names.size)
    source = _select_index([source], df, mapping[axis])
    source = _index_converter(source, index)
    if target is None:
        if position == "after":
            target = np.array([names.size])
        else:
            target = np.array([0])
    else:
        target = _select_index([target], df, mapping[axis])
        target = _index_converter(target, index)
    index = np.delete(index, source)

    if position == "before":
        position = index.searchsorted(target[0])
    else:
        position = index.searchsorted(target[-1]) + 1
    start = index[:position]
    end = index[position:]
    position = np.concatenate([start, source, end])

    return df.iloc(axis=axis)[position]



"""Tests for test helper functions."""

from helpers import running_on_ci


def test_running_on_ci_local(monkeypatch):
    """Test running_on_ci run on local machine returns False."""
    monkeypatch.delenv("JANITOR_CI_MACHINE", raising=False)
    assert running_on_ci() is False


def test_running_on_ci_ci(monkeypatch):
    """Test running_on_ci run on CI machine returns True."""
    monkeypatch.setenv("JANITOR_CI_MACHINE", "1")
    assert running_on_ci() is True



# helper function for a single join

import pandas as pd

from janitor.functions._conditional_join._greater_than_indices import (
    _greater_than_indices,
)
from janitor.functions._conditional_join._helpers import (
    greater_than_join_types,
    less_than_join_types,
)
from janitor.functions._conditional_join._less_than_indices import (
    _less_than_indices,
)
from janitor.functions._conditional_join._not_equal_indices import (
    _not_equal_indices,
)


def _single_join(
    df: pd.DataFrame,
    right: pd.DataFrame,
    condition: tuple,
    keep: str,
    return_matching_indices: bool,
) -> dict:
    """
    Compute indices for a single join
    """
    left_on, right_on, op = condition
    if op in less_than_join_types:
        return _less_than_indices(
            left=df[left_on],
            right=right[right_on],
            strict=op == "<",
            keep=keep,
            return_matching_indices=return_matching_indices,
        )
    if op in greater_than_join_types:
        return _greater_than_indices(
            left=df[left_on],
            right=right[right_on],
            strict=op == ">",
            keep=keep,
            return_matching_indices=return_matching_indices,
        )
    if op == "!=":
        return _not_equal_indices(
            left=df[left_on],
            right=right[right_on],
            keep=keep,
        )



from .clean_names import clean_names, make_clean_names
from .complete import complete, expand
from .dates_to_polars import convert_excel_date, convert_matlab_date
from .pivot_longer import pivot_longer, pivot_longer_spec
from .row_to_names import row_to_names

__all__ = [
    "pivot_longer_spec",
    "pivot_longer",
    "clean_names",
    "make_clean_names",
    "row_to_names",
    "expand",
    "complete",
    "convert_excel_date",
    "convert_matlab_date",
]



"""Implementation of drop_constant_columns."""

import pandas as pd
import pandas_flavor as pf


@pf.register_dataframe_method
def drop_constant_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Finds and drops the constant columns from a Pandas DataFrame.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> data_dict = {
        ...     "a": [1, 1, 1],
        ...     "b": [1, 2, 3],
        ...     "c": [1, 1, 1],
        ...     "d": ["rabbit", "leopard", "lion"],
        ...     "e": ["Cambridge", "Shanghai", "Basel"],
        ... }
        >>> df = pd.DataFrame(data_dict)
        >>> df
           a  b  c        d          e
        0  1  1  1   rabbit  Cambridge
        1  1  2  1  leopard   Shanghai
        2  1  3  1     lion      Basel
        >>> df.drop_constant_columns()
           b        d          e
        0  1   rabbit  Cambridge
        1  2  leopard   Shanghai
        2  3     lion      Basel

    Args:
        df: Input Pandas DataFrame

    Returns:
        The Pandas DataFrame with the constant columns dropped.
    """
    return df.loc[:, df.nunique().ne(1)]



import datetime as dt
import warnings
from functools import reduce
from typing import Any, Dict, Hashable, Iterable, List, Optional

import numpy as np
import pandas as pd
import pandas_flavor as pf

from janitor.utils import (
    deprecated_alias,
    find_stack_level,
    refactored_function,
)

warnings.simplefilter("always", DeprecationWarning)


@pf.register_dataframe_method
@deprecated_alias(column="column_name")
def filter_string(
    df: pd.DataFrame,
    column_name: Hashable,
    search_string: str,
    complement: bool = False,
    case: bool = True,
    flags: int = 0,
    na: Any = None,
    regex: bool = True,
) -> pd.DataFrame:
    """Filter a string-based column according to whether it contains a substring.

    This is super sugary syntax that builds on top of `pandas.Series.str.contains`.
    It is meant to be the method-chaining equivalent of the following:

    ```python
    df = df[df[column_name].str.contains(search_string)]]
    ```

    This method does not mutate the original DataFrame.

    Examples:
        Retain rows whose column values contain a particular substring.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": range(3, 6), "b": ["bear", "peeL", "sail"]})
        >>> df
           a     b
        0  3  bear
        1  4  peeL
        2  5  sail
        >>> df.filter_string(column_name="b", search_string="ee")
           a     b
        1  4  peeL
        >>> df.filter_string(column_name="b", search_string="L", case=False)
           a     b
        1  4  peeL
        2  5  sail

        Filter names does not contain `'.'` (disable regex mode).

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.Series(["JoseChen", "Brian.Salvi"], name="Name").to_frame()
        >>> df
                  Name
        0     JoseChen
        1  Brian.Salvi
        >>> df.filter_string(
        ...     column_name="Name", search_string=".", regex=False, complement=True
        ... )
               Name
        0  JoseChen

    Args:
        df: A pandas DataFrame.
        column_name: The column to filter. The column should contain strings.
        search_string: A regex pattern or a (sub-)string to search.
        complement: Whether to return the complement of the filter or not. If
            set to True, then the rows for which the string search fails are retained
            instead.
        case: If True, case sensitive.
        flags: Flags to pass through to the re module, e.g. re.IGNORECASE.
        na: Fill value for missing values. The default depends on dtype of
            the array. For object-dtype, `numpy.nan` is used. For `StringDtype`,
            `pandas.NA` is used.
        regex: If True, assumes `search_string` is a regular expression. If False,
            treats the `search_string` as a literal string.

    Returns:
        A filtered pandas DataFrame.
    """  # noqa: E501

    criteria = df[column_name].str.contains(
        pat=search_string,
        case=case,
        flags=flags,
        na=na,
        regex=regex,
    )

    if complement:
        return df[~criteria]

    return df[criteria]


@pf.register_dataframe_method
@refactored_function(
    message=(
        "This function will be deprecated in a 1.x release. "
        "Please use `pd.DataFrame.query` instead."
    )
)
def filter_on(
    df: pd.DataFrame,
    criteria: str,
    complement: bool = False,
) -> pd.DataFrame:
    """Return a dataframe filtered on a particular criteria.

    This method does not mutate the original DataFrame.

    This is super-sugary syntax that wraps the pandas `.query()` API, enabling
    users to use strings to quickly specify filters for filtering their
    dataframe. The intent is that `filter_on` as a verb better matches the
    intent of a pandas user than the verb `query`.

    This is intended to be the method-chaining equivalent of the following:

    ```python
    df = df[df["score"] < 3]
    ```

    !!!note

        This function will be deprecated in a 1.x release.
        Please use `pd.DataFrame.query` instead.


    Examples:
        Filter students who failed an exam (scored less than 50).

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "student_id": ["S1", "S2", "S3"],
        ...         "score": [40, 60, 85],
        ...     }
        ... )
        >>> df
          student_id  score
        0         S1     40
        1         S2     60
        2         S3     85
        >>> df.filter_on("score < 50", complement=False)
          student_id  score
        0         S1     40

    Credit to Brant Peterson for the name.

    Args:
        df: A pandas DataFrame.
        criteria: A filtering criteria that returns an array or Series of
            booleans, on which pandas can filter on.
        complement: Whether to return the complement of the filter or not.
            If set to True, then the rows for which the criteria is False are
            retained instead.

    Returns:
        A filtered pandas DataFrame.
    """

    warnings.warn(
        "This function will be deprecated in a 1.x release. "
        "Kindly use `pd.DataFrame.query` instead.",
        DeprecationWarning,
        stacklevel=find_stack_level(),
    )

    if complement:
        return df.query(f"not ({criteria})")
    return df.query(criteria)


@pf.register_dataframe_method
@deprecated_alias(column="column_name", start="start_date", end="end_date")
def filter_date(
    df: pd.DataFrame,
    column_name: Hashable,
    start_date: Optional[dt.date] = None,
    end_date: Optional[dt.date] = None,
    years: Optional[List] = None,
    months: Optional[List] = None,
    days: Optional[List] = None,
    column_date_options: Optional[Dict] = None,
    format: Optional[str] = None,  # skipcq: PYL-W0622
) -> pd.DataFrame:
    """Filter a date-based column based on certain criteria.

    This method does not mutate the original DataFrame.

    Dates may be finicky and this function builds on top of the *magic* from
    the pandas `to_datetime` function that is able to parse dates well.

    Additional options to parse the date type of your column may be found at
    the official pandas [documentation][datetime].

    [datetime]: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.to_datetime.html

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "a": range(5, 9),
        ...         "dt": ["2021-11-12", "2021-12-15", "2022-01-03", "2022-01-09"],
        ...     }
        ... )
        >>> df
           a          dt
        0  5  2021-11-12
        1  6  2021-12-15
        2  7  2022-01-03
        3  8  2022-01-09
        >>> df.filter_date("dt", start_date="2021-12-01", end_date="2022-01-05")
           a         dt
        1  6 2021-12-15
        2  7 2022-01-03
        >>> df.filter_date("dt", years=[2021], months=[12])
           a         dt
        1  6 2021-12-15

    !!!note

        This method will cast your column to a Timestamp!

    !!!note

        This only affects the format of the `start_date` and `end_date`
        parameters. If there's an issue with the format of the DataFrame being
        parsed, you would pass `{'format': your_format}` to `column_date_options`.

    Args:
        df: The dataframe to filter on.
        column_name: The column which to apply the fraction transformation.
        start_date: The beginning date to use to filter the DataFrame.
        end_date: The end date to use to filter the DataFrame.
        years: The years to use to filter the DataFrame.
        months: The months to use to filter the DataFrame.
        days: The days to use to filter the DataFrame.
        column_date_options: Special options to use when parsing the date
            column in the original DataFrame. The options may be found at the
            official Pandas documentation.
        format: If you're using a format for `start_date` or `end_date`
            that is not recognized natively by pandas' `to_datetime` function, you
            may supply the format yourself. Python date and time formats may be
            found [here](http://strftime.org/).

    Returns:
        A filtered pandas DataFrame.
    """  # noqa: E501

    def _date_filter_conditions(conditions):
        """Taken from: https://stackoverflow.com/a/13616382."""
        return reduce(np.logical_and, conditions)

    if column_date_options is None:
        column_date_options = {}
    df[column_name] = pd.to_datetime(df[column_name], **column_date_options)

    _filter_list = []

    if start_date:
        start_date = pd.to_datetime(start_date, format=format)
        _filter_list.append(df[column_name] >= start_date)

    if end_date:
        end_date = pd.to_datetime(end_date, format=format)
        _filter_list.append(df[column_name] <= end_date)

    if years:
        _filter_list.append(df[column_name].dt.year.isin(years))

    if months:
        _filter_list.append(df[column_name].dt.month.isin(months))

    if days:
        _filter_list.append(df[column_name].dt.day.isin(days))

    if start_date and end_date and start_date > end_date:
        warnings.warn(
            f"Your start date of {start_date} is after your end date of "
            f"{end_date}. Is this intended?"
        )

    return df.loc[_date_filter_conditions(_filter_list), :]


@pf.register_dataframe_method
@refactored_function(
    message=(
        "This function will be deprecated in a 1.x release. "
        "Please use `pd.DataFrame.query` or `pd.DataFrame.isin` instead."
    )
)
@deprecated_alias(column="column_name")
def filter_column_isin(
    df: pd.DataFrame,
    column_name: Hashable,
    iterable: Iterable,
    complement: bool = False,
) -> pd.DataFrame:
    """Filter a dataframe for values in a column that exist in the given iterable.

    This method does not mutate the original DataFrame.

    Assumes exact matching; fuzzy matching not implemented.

    Examples:
        Filter the dataframe to retain rows for which `names`
        are exactly `James` or `John`.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"names": ["Jane", "Jeremy", "John"], "foo": list("xyz")})
        >>> df
            names foo
        0    Jane   x
        1  Jeremy   y
        2    John   z
        >>> df.filter_column_isin(column_name="names", iterable=["James", "John"])
          names foo
        2  John   z

        This is the method-chaining alternative to:

        ```python
        df = df[df["names"].isin(["James", "John"])]
        ```

        If `complement=True`, then we will only get rows for which the names
        are neither `James` nor `John`.

    Args:
        df: A pandas DataFrame.
        column_name: The column on which to filter.
        iterable: An iterable. Could be a list, tuple, another pandas
            Series.
        complement: Whether to return the complement of the selection or
            not.

    Raises:
        ValueError: If `iterable` does not have a length of `1`
            or greater.

    Returns:
        A filtered pandas DataFrame.
    """  # noqa: E501

    warnings.warn(
        "This function will be deprecated in a 1.x release. "
        "Kindly use `pd.DataFrame.query` or `pd.DataFrame.isin` instead.",
        DeprecationWarning,
        stacklevel=find_stack_level(),
    )

    if len(iterable) == 0:
        raise ValueError(
            "`iterable` kwarg must be given an iterable of length 1 or greater."
        )
    criteria = df[column_name].isin(iterable)

    if complement:
        return df[~criteria]
    return df[criteria]



"""Implementation of the `join_apply` function"""

from typing import Callable

import pandas as pd
import pandas_flavor as pf


@pf.register_dataframe_method
def join_apply(
    df: pd.DataFrame,
    func: Callable,
    new_column_name: str,
) -> pd.DataFrame:
    """Join the result of applying a function across dataframe rows.

    This method does not mutate the original DataFrame.

    This is a convenience function that allows us to apply arbitrary functions
    that take any combination of information from any of the columns. The only
    requirement is that the function signature takes in a row from the
    DataFrame.

    Examples:
        Sum the result of two columns into a new column.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": [1, 2, 3], "b": [2, 3, 4]})
        >>> df
           a  b
        0  1  2
        1  2  3
        2  3  4
        >>> df.join_apply(
        ...     func=lambda x: 2 * x["a"] + x["b"],
        ...     new_column_name="2a+b",
        ... )
           a  b  2a+b
        0  1  2     4
        1  2  3     7
        2  3  4    10

        Incorporating conditionals in `func`.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": [1, 2, 3], "b": [20, 30, 40]})
        >>> df
           a   b
        0  1  20
        1  2  30
        2  3  40
        >>> def take_a_if_even(x):
        ...     if x["a"] % 2 == 0:
        ...         return x["a"]
        ...     else:
        ...         return x["b"]
        >>> df.join_apply(take_a_if_even, "a_if_even")
           a   b  a_if_even
        0  1  20         20
        1  2  30          2
        2  3  40         40

    Args:
        df: A pandas DataFrame.
        func: A function that is applied elementwise across all rows of the
            DataFrame.
        new_column_name: Name of the resulting column.

    Returns:
        A pandas DataFrame with new column appended.
    """  # noqa: E501
    df = df.copy().join(df.apply(func, axis=1).rename(new_column_name))
    return df



from typing import Hashable, List

import pandas as pd
import pandas_flavor as pf

from janitor.errors import JanitorError
from janitor.utils import deprecated_alias


@pf.register_dataframe_method
@deprecated_alias(columns="column_names")
def concatenate_columns(
    df: pd.DataFrame,
    column_names: List[Hashable],
    new_column_name: Hashable,
    sep: str = "-",
    ignore_empty: bool = True,
) -> pd.DataFrame:
    """Concatenates the set of columns into a single column.

    Used to quickly generate an index based on a group of columns.

    This method mutates the original DataFrame.

    Examples:
        Concatenate two columns row-wise.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": [1, 3, 5], "b": list("xyz")})
        >>> df
           a  b
        0  1  x
        1  3  y
        2  5  z
        >>> df.concatenate_columns(
        ...     column_names=["a", "b"],
        ...     new_column_name="m",
        ... )
           a  b    m
        0  1  x  1-x
        1  3  y  3-y
        2  5  z  5-z

    Args:
        df: A pandas DataFrame.
        column_names: A list of columns to concatenate together.
        new_column_name: The name of the new column.
        sep: The separator between each column's data.
        ignore_empty: Ignore null values if exists.

    Raises:
        JanitorError: If at least two columns are not provided
            within `column_names`.

    Returns:
        A pandas DataFrame with concatenated columns.
    """
    if len(column_names) < 2:
        raise JanitorError("At least two columns must be specified")

    df[new_column_name] = df[column_names].astype(str).fillna("").agg(sep.join, axis=1)

    if ignore_empty:

        def remove_empty_string(x):
            """Ignore empty/null string values from the concatenated output."""
            return sep.join(x for x in x.split(sep) if x)

        df[new_column_name] = df[new_column_name].transform(remove_empty_string)

    return df



"""Implementation source for `then`."""

from typing import Callable

import pandas as pd
import pandas_flavor as pf

from janitor.utils import refactored_function


@pf.register_dataframe_method
@refactored_function(
    message="This function will be deprecated in a 1.x release. "
    "Kindly use `pd.DataFrame.pipe` instead."
)
def then(df: pd.DataFrame, func: Callable) -> pd.DataFrame:
    """Add an arbitrary function to run in the `pyjanitor` method chain.

    This method does not mutate the original DataFrame.

    !!!note

        This function will be deprecated in a 1.x release.
        Please use `pd.DataFrame.pipe` instead.

    Examples:
        A trivial example using a lambda `func`.

        >>> import pandas as pd
        >>> import janitor
        >>> (pd.DataFrame({"a": [1, 2, 3], "b": [7, 8, 9]}).then(lambda df: df * 2))
           a   b
        0  2  14
        1  4  16
        2  6  18

    Args:
        df: A pandas DataFrame.
        func: A function you would like to run in the method chain.
            It should take one parameter and return one parameter, each being
            the DataFrame object. After that, do whatever you want in the
            middle. Go crazy.

    Returns:
        A pandas DataFrame.
    """
    df = func(df)
    return df



import numpy as np
import pandas as pd

from janitor.functions._conditional_join import (
    _dual_non_equi,
    _helpers,
)


def _get_indices(
    df: pd.DataFrame,
    right: pd.DataFrame,
    mapping: dict,
    return_matching_indices: bool,
    keep: str,
):
    outcome = _dual_non_equi._get_indices(
        df=df,
        right=right,
        first_condition=mapping["ge_gt"],
        second_condition=mapping["le_lt"],
    )
    empty_array = np.array([], dtype=np.intp)
    if outcome is None:
        return {
            "left_index": empty_array,
            "right_index": empty_array,
        }
    rest = []
    rest.extend(mapping["le_or_ge"])
    rest.extend(mapping["equals"])
    rest.extend(mapping["not_equals"])
    rest = [entry for entry in rest if entry]
    if return_matching_indices and not rest:
        starts, ends = _dual_non_equi._build_starts_and_ends(
            counts=outcome["counts_array"]
        )
        outcome["starts"] = starts
        outcome["ends"] = ends
        return outcome
    if not rest:
        if keep == "all":
            starts = None
            ends = None
        else:
            starts, ends = _dual_non_equi._build_starts_and_ends(
                counts=outcome["counts_array"]
            )
        return _helpers._build_indices_positions(
            left_index=outcome["left_index"],
            right_index=outcome["right_index"],
            positions=outcome["positions"],
            starts=starts,
            ends=ends,
            counts_array=outcome["counts_array"],
            total=outcome["total"],
            keep=keep,
        )
    starts, ends = _dual_non_equi._build_starts_and_ends(counts=outcome["counts_array"])
    out = _helpers._get_positive_matches_conditions_posns(
        df=df,
        right=right,
        conditions=rest,
        left_index=outcome["left_index"],
        right_index=outcome["right_index"],
        positions=outcome["positions"],
        starts=starts,
        ends=ends,
    )
    if out is None:
        return {
            "left_index": empty_array,
            "right_index": empty_array,
        }
    if return_matching_indices:
        out["starts"] = starts
        out["ends"] = ends
        out = outcome | out
        return out
    return _helpers._build_indices_positions(
        left_index=outcome["left_index"],
        right_index=outcome["right_index"],
        positions=out["positions"],
        starts=starts,
        ends=ends,
        counts_array=out["counts_array"],
        total=out["total"],
        keep=keep,
    )



"""Implementation of `round_half_up` and `signif_half_up`."""

from __future__ import annotations

from typing import Callable, Tuple

import numpy as np
import pandas as pd


def _prepare_numeric_input(
    x,
) -> Tuple[np.ndarray, Callable[[np.ndarray], object]]:
    """Convert input to a numeric array and build a wrapper for output."""
    if isinstance(x, pd.Series):
        arr = x.to_numpy(dtype=float)

        def wrap(result: np.ndarray):
            return pd.Series(result, index=x.index, name=x.name)

        return arr, wrap

    arr = np.asarray(x, dtype=float)
    is_scalar = arr.ndim == 0
    arr = np.atleast_1d(arr)

    def wrap(result: np.ndarray):
        if is_scalar:
            return result[0].item()
        return result

    return arr, wrap


def round_half_up(x, digits: int = 0):
    """Round values using "half up" rounding (Excel-style).

    This differs from Python's built-in `round`, which uses bankers rounding.

    Examples:
        >>> import janitor
        >>> janitor.round_half_up(12.5)
        13.0
        >>> janitor.round_half_up(1.125, digits=2)
        1.13
        >>> janitor.round_half_up(-0.5)
        -1.0

    Args:
        x: Numeric scalar or array-like.
        digits: Number of decimal digits to round to.

    Returns:
        Rounded value(s).
    """
    arr, wrap = _prepare_numeric_input(x)
    factor = 10**digits
    posneg = np.sign(arr)
    z = np.abs(arr) * factor
    z = z + 0.5 + np.sqrt(np.finfo(float).eps)
    z = np.trunc(z)
    z = z / factor
    z = z * posneg
    return wrap(z)


def signif_half_up(x, digits: int = 6):
    """Round values to the specified number of significant digits.

    Uses "half up" rounding for midpoint values.

    Examples:
        >>> import janitor
        >>> janitor.signif_half_up(12.5, digits=2)
        13.0
        >>> janitor.signif_half_up(1.125, digits=3)
        1.13
        >>> janitor.signif_half_up(-2.5, digits=1)
        -3.0

    Args:
        x: Numeric scalar or array-like.
        digits: Number of significant digits.

    Returns:
        Rounded value(s).
    """
    arr, wrap = _prepare_numeric_input(x)
    z = arr.astype(float, copy=True)
    mask = (z != 0) & np.isfinite(z)
    if mask.any():
        y = np.zeros_like(z)
        y[mask] = 10 ** (digits - np.ceil(np.log10(np.abs(z[mask]))))
        z[mask] = round_half_up(z[mask] * y[mask]) / y[mask]
    return wrap(z)



from __future__ import annotations

import janitor_rs
import numpy as np
import pandas as pd

from janitor.functions._conditional_join import (
    _helpers,
)


def _get_indices(
    df: pd.DataFrame,
    right: pd.DataFrame,
    mapping: dict,
):
    empty_array = np.array([], dtype=np.intp)
    l_cols = []
    r_cols = []
    for left_col, right_col, _ in mapping["equals"]:
        l_cols.append(df[left_col]._values)
        r_cols.append(right[right_col]._values)
    if len(l_cols) > 1:
        l_cols = pd.MultiIndex.from_arrays(l_cols)
        r_cols = pd.MultiIndex.from_arrays(r_cols)
    else:
        l_cols = pd.Index(l_cols[0])
        r_cols = pd.Index(r_cols[0])
    left_index = df.index._values
    right_index = right.index._values
    indexers = r_cols.get_indexer(l_cols)
    booleans = indexers != -1
    if not booleans.any():
        return {
            "left_index": empty_array,
            "right_index": empty_array,
        }
    if not booleans.all() and not any(
        (
            mapping["le_or_ge"],
            mapping["le_lt"],
            mapping["ge_gt"],
            mapping["not_equals"],
        )
    ):
        indexers = indexers[booleans]
        left_index = left_index[booleans]
        right_index = right_index[indexers]
        return {
            "left_index": left_index,
            "right_index": right_index,
        }
    if not any(
        (
            mapping["le_or_ge"],
            mapping["le_lt"],
            mapping["ge_gt"],
            mapping["not_equals"],
        )
    ):
        right_index = right_index[indexers]
        return {
            "left_index": left_index,
            "right_index": right_index,
        }
    rest = mapping["le_or_ge"]
    rest.append(mapping["le_lt"])
    rest.append(mapping["ge_gt"])
    rest.extend(mapping["not_equals"])
    rest = filter(None, rest)
    rest = dict.fromkeys(rest)
    outcome = _helpers._update_positions_no_range_(
        df=df, right=right, conditions=rest, positions=indexers
    )
    if outcome is None:
        return {
            "left_index": empty_array,
            "right_index": empty_array,
        }
    left_index = janitor_rs.index_trim_positions(
        index=left_index,
        positions=outcome["positions"],
        length=outcome["total"],
    )
    right_index = janitor_rs.build_positional_index(
        index=right_index,
        positions=outcome["positions"],
        length=outcome["total"],
    )
    return {
        "left_index": left_index,
        "right_index": right_index,
    }



import janitor_rs
import numpy as np
import pandas as pd

from janitor.functions._conditional_join import (
    _binary_search,
    _helpers,
)

# description below is based on multiple non-equi joins (>/>=/</<=)
# implementation is based on the algorithm described in this paper -
# https://www.scitepress.org/papers/2018/68268/68268.pdf

# summary:
# get regions for first and second conditions in the pair
# (l_col1, r_col1, op1), (l_col2, r_col2, op2)
# the idea is that r_col1 should always be ahead of the
# appropriate value from lcol1; same applies to l_col2 & r_col2.
# if the operator is in less than join types
# the l_col should be in ascending order
# if in greater than join types, l_col should be
# in descending order
# Example :
#     df1:
#    id  value_1
# 0   1        2
# 1   1        5
# 2   1        7
# 3   2        1
# 4   2        3
# 5   3        4
#
#
#  df2:
#    id  value_2A  value_2B
# 0   1         0         1
# 1   1         3         5
# 2   1         7         9
# 3   1        12        15
# 4   2         0         1
# 5   2         2         4
# 6   2         3         6
# 7   3         1         3
#
#
# ('value_1', 'value_2A','>'), ('value_1', 'value_2B', '<')
# for the first pair, since op is greater than
# 'value_1' is sorted in descending order
#  our pairing should be :
# value  source      region number
# 12   value_2A       0
# 7    value_2A       1
# 7    value_1        2
# 5    value_1        2
# 4    value_1        2
# 3    value_2A       2
# 3    value_2A       2
# 3    value_1        3
# 2    value_2A       3
# 2    value_1        4
# 1    value_2A       4
# 1    value_1        5
# 0    value_2A       5
# 0    value_2A       5
#
# note that 7 for value_2A is not matched with 7 of value_1
# because it is >, not >=, hence the different region numbers
# looking at the output above, we can safely discard regions 0 and 1
# since they do not have any matches with value_1
# for the second pair, since op is <, value_1 is sorted
# in ascending order, and our pairing should be:
#   value    source    region number
#     1    value_2B       0
#     1    value_2B       1
#     1    value_1        2
#     2    value_1        2
#     3    value_2B       2
#     3    value_1        3
#     4    value_2B       3
#     4    value_1        4
#     5    value_2B       4
#     5    value_1        5
#     6    value_2B       5
#     7    value_1        6
#     9    value_2B       6
#     15   value_2B       6
#
# from the above we can safely discard regions 0 and 1, since there are
# no matches with value_1 ... note that the index for regions 0 and 1
# coincide with the index for region 5 values in value_2A(0, 0);
# as such those regions will be discarded.
# Similarly, the index for regions 0 and 1 of value_2A(12, 7)
# coincide with the index for regions 6 for value_2B(9, 15);
# these will be discarded as well.
# let's create a table of the regions, paired with the index
#
#
#  value_1 :
###############################################
# index-->  2  1  5  4  0  3
# pair1-->  2  2  2  3  4  5
# pair2-->  6  5  4  3  2  2
###############################################
#
#
# value_2A, value_2B
##############################################
# index --> 1  6  5  7
# pair1 --> 2  2  3  4
# pair2 --> 4  5  3  2
##############################################
#
# To find matching indices, the regions from value_1 must be less than
# or equal to the regions in value_2A/2B.
# pair1 <= pair1 and pair2 <= pair2
# Starting from the highest region in value_1
# 5 in pair1 is not less than any in value_2A/2B, so we discard
# 4 in pair1 is matched to 4 in pair1 of value_2A/2B
# we look at the equivalent value in pair2 for 4, which is 2
# 2 matches 2 in pair 2, so we have a match -> (0, 7)
# 3 in pair 1 from value_1 matches 3 and 4 in pair1 for value_2A/2B
# next we compare the equivalent value from pair2, which is 3
# 3 matches only 3 in value_2A/2B, so our only match is  -> (4, 5)
# next is 2 (we have 3 2s in value_1 for pair1)
# they all match 2, 2, 3, 4 in pair1 of value_2A/2B
# compare the first equivalent in pair2 -> 4
# 4 matches only 4, 5 in pair2 of value_2A/2B
# ->(5, 1), (5, 6)
# the next equivalent is -> 5
# 5 matches only 5 in pair2 of value_2A/2B
# -> (1, 6)
# the last equivalent is -> 6
# 6 has no match in pair2 of value_2A/2B, so we discard
# our final matching indices for the left and right pairs
#########################################################
# left_index      right_index
#     0              7
#     4              5
#     5              1
#     5              6
#     1              6
########################################################
# and if we index the dataframes,
# we should get the output below:
#################################
#    value_1  value_2A  value_2B
# 0        2         1         3
# 1        5         3         6
# 2        3         2         4
# 3        4         3         5
# 4        4         3         6
################################


def _get_indices(
    df: pd.DataFrame, right: pd.DataFrame, first_condition, second_condition
) -> dict:
    """
    Compute indices for a dual non-equi join
    """
    (left_on, right_on, op) = first_condition
    left_col = df[left_on]
    # sorting is done here to enable easy region filtering later on
    # by starting from the highest region
    left_col, _ = _helpers._sort_if_not_monotonic(series=left_col)
    right_col = right[right_on]
    outcome = _build_regions(left=left_col, right=right_col, op=op)
    if outcome is None:
        return None
    l1_index, l1_region, r1_index, r1_region = outcome
    (left_on, right_on, op) = second_condition
    left_col = df.loc[l1_index, left_on]
    right_col = right.loc[r1_index, right_on]
    outcome = _build_regions(left=left_col, right=right_col, op=op)
    if outcome is None:
        return None
    l2_index, l2_region, r2_index, r2_region = outcome
    outcome = _align_regions(
        left_index=l1_index,
        right_index=l2_index,
        left_region=l1_region,
        right_region=l2_region,
    )
    if outcome is None:
        return None
    l_index, l1_region, l2_region = outcome
    outcome = _align_regions(
        left_index=r1_index,
        right_index=r2_index,
        left_region=r1_region,
        right_region=r2_region,
    )
    if outcome is None:
        return None
    r_index, r1_region, r2_region = outcome
    # flip to ensure region filtering
    # starts from the highest search position
    if not pd.Index(l1_region).is_monotonic_decreasing:
        l1_region = l1_region[::-1]
        l_index = l_index[::-1]
        l2_region = l2_region[::-1]
    # due to the approach in _build_regions and _align_regions,
    # r1_region is guaranteed to be sorted
    search_indices = r1_region.searchsorted(l1_region)
    # keep only rows where
    # l_region <= r_region
    booleans = search_indices == r1_region.size
    if booleans.all():
        return None
    if booleans.any():
        booleans = ~booleans
        l_index = l_index[booleans]
        l2_region = l2_region[booleans]
        search_indices = search_indices[booleans]
    # no point searching if l2_region > r2_region
    max_region = r2_region.max()
    booleans = l2_region > max_region
    if booleans.all():
        return None
    # keep only rows where
    # l_region <= r_region
    if booleans.any():
        booleans = ~booleans
        l_index = l_index[booleans]
        l2_region = l2_region[booleans]
        search_indices = search_indices[booleans]
    # at this point there cant be any -1s
    # this is the base stage
    counts_array, positions, total = janitor_rs.get_positions_where_left_le_right(
        left=l2_region,
        right=r2_region,
        starts=search_indices,
        max_right=max_region,
    )
    if not total:
        return None
    return {
        "left_index": l_index,
        "right_index": r_index,
        "positions": positions,
        "counts_array": counts_array,
        "total": total,
    }


def _build_regions(left: pd.Series, right: pd.Series, op: str) -> tuple:
    """Compute regions"""
    right, _ = _helpers._sort_if_not_monotonic(series=right)
    left_array = _helpers._convert_array_to_numpy(array=left._values)
    right_array = _helpers._convert_array_to_numpy(array=right._values)
    if op == "<":
        outcome = _binary_search._binary_search_lt_first(
            left=left_array, right=right_array, left_index=left.index._values
        )
    elif op == "<=":
        outcome = _binary_search._binary_search_le_first(
            left=left_array, right=right_array, left_index=left.index._values
        )
    elif op == ">":
        outcome = _binary_search._binary_search_gt_regions(
            left=left_array, right=right_array, left_index=left.index._values
        )
    else:
        outcome = _binary_search._binary_search_ge_regions(
            left=left_array, right=right_array, left_index=left.index._values
        )
    if outcome is None:
        return None
    # computation here is to ensure alignment
    # since right should be in descending order
    # for >/>=
    if op in _helpers.greater_than_join_types:
        right_index = right.index._values[::-1]
    else:
        right_index = right.index._values
    left_index, search_indices = outcome
    right_region = np.zeros(right_index.size, dtype=np.int64)
    right_region[search_indices] = 1
    right_region[0] -= 1
    right_region = right_region.cumsum()
    left_region = right_region[search_indices]
    # exclude regions that definitely do not have a match
    booleans = right_region > -1
    if not booleans.all():
        right_region = right_region[booleans]
        right_index = right_index[booleans]
    return left_index, left_region, right_index, right_region


def _align_regions(
    left_index: np.ndarray,
    right_index: np.ndarray,
    left_region: np.ndarray,
    right_region: np.ndarray,
) -> tuple:
    "Ensure regions are aligned"
    # right_index is always going to be <= left_index
    # based on previous computations,
    # right_index will always be a subset of left_index
    # as such, right_index becomes the control point
    positions = pd.Index(right_index).get_indexer(left_index)
    booleans = positions > -1
    if not booleans.any():
        return None
    left_index = left_index[booleans]
    left_region = left_region[booleans]
    positions = positions[booleans]
    right_index = right_index[positions]
    right_region = right_region[positions]
    return left_index, left_region, right_region


def _build_starts_and_ends(counts: np.ndarray) -> tuple:
    ends = counts.cumsum()
    starts = np.empty(counts.size, dtype=np.int64)
    starts[0] = 0
    starts[1:] = ends[:-1]
    return starts, ends



"""Implementation of the `factorize_columns` function"""

from typing import Any, Hashable, Iterable, Union

import pandas as pd
import pandas_flavor as pf

from janitor.functions.utils import _factorize


@pf.register_dataframe_method
def factorize_columns(
    df: pd.DataFrame,
    column_names: Union[str, Iterable[str], Hashable],
    suffix: str = "_enc",
    **kwargs: Any,
) -> pd.DataFrame:
    """Converts labels into numerical data.

    This method will create a new column with the string `_enc` appended
    after the original column's name.
    This can be overridden with the suffix parameter.

    Internally, this method uses pandas `factorize` method.
    It takes in an optional suffix and keyword arguments also.
    An empty string as suffix will override the existing column.

    This method does not mutate the original DataFrame.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "foo": ["b", "b", "a", "c", "b"],
        ...         "bar": range(4, 9),
        ...     }
        ... )
        >>> df
          foo  bar
        0   b    4
        1   b    5
        2   a    6
        3   c    7
        4   b    8
        >>> df.factorize_columns(column_names="foo")
          foo  bar  foo_enc
        0   b    4        0
        1   b    5        0
        2   a    6        1
        3   c    7        2
        4   b    8        0

    Args:
        df: The pandas DataFrame object.
        column_names: A column name or an iterable (list or tuple) of
            column names.
        suffix: Suffix to be used for the new column.
            An empty string suffix means, it will override the existing column.
        **kwargs: Keyword arguments. It takes any of the keyword arguments,
            which the pandas factorize method takes like `sort`, `na_sentinel`,
            `size_hint`.

    Returns:
        A pandas DataFrame.
    """
    df = _factorize(df.copy(), column_names, suffix, **kwargs)
    return df



"""Implementation source for `process_text`."""

import inspect
from typing import Any

import pandas as pd
import pandas_flavor as pf

from janitor.utils import (
    check,
    check_column,
    deprecated_alias,
    refactored_function,
)


@pf.register_dataframe_method
@refactored_function(
    message=(
        "This function will be deprecated in a 1.x release. "
        "Please use `jn.transform_columns` instead."
    )
)
@deprecated_alias(column="column_name")
def process_text(
    df: pd.DataFrame,
    column_name: str,
    string_function: str,
    **kwargs: Any,
) -> pd.DataFrame:
    """Apply a Pandas string method to an existing column.

    This function aims to make string cleaning easy, while chaining,
    by simply passing the string method name,
    along with keyword arguments, if any, to the function.

    This modifies an existing column; it does not create a new column;
    new columns can be created via pyjanitor's
    [`transform_columns`][janitor.functions.transform_columns.transform_columns].

    A list of all the string methods in Pandas can be accessed [here](https://pandas.pydata.org/docs/user_guide/text.html#method-summary).

    !!!note

        This function will be deprecated in a 1.x release.
        Please use [`jn.transform_column`][janitor.functions.transform_columns.transform_column]
        instead.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> import re
        >>> df = pd.DataFrame(
        ...     {"text": ["Ragnar", "sammywemmy", "ginger"], "code": [1, 2, 3]}
        ... )
        >>> df
                 text  code
        0      Ragnar     1
        1  sammywemmy     2
        2      ginger     3
        >>> df.process_text(column_name="text", string_function="lower")
                 text  code
        0      ragnar     1
        1  sammywemmy     2
        2      ginger     3

        For string methods with parameters, simply pass the keyword arguments:

        >>> df.process_text(
        ...     column_name="text",
        ...     string_function="extract",
        ...     pat=r"(ag)",
        ...     expand=False,
        ...     flags=re.IGNORECASE,
        ... )
          text  code
        0   ag     1
        1  NaN     2
        2  NaN     3

    Args:
        df: A pandas DataFrame.
        column_name: String column to be operated on.
        string_function: pandas string method to be applied.
        **kwargs: Keyword arguments for parameters of the `string_function`.

    Raises:
        KeyError: If `string_function` is not a Pandas string method.
        ValueError: If the text function returns a DataFrame, instead of a Series.

    Returns:
        A pandas DataFrame with modified column.
    """  # noqa: E501

    check("column_name", column_name, [str])
    check("string_function", string_function, [str])
    check_column(df, [column_name])

    pandas_string_methods = [
        func.__name__
        for _, func in inspect.getmembers(pd.Series.str, inspect.isfunction)
        if not func.__name__.startswith("_")
    ]

    if string_function not in pandas_string_methods:
        raise KeyError(f"{string_function} is not a Pandas string method.")

    result = getattr(df[column_name].str, string_function)(**kwargs)

    if isinstance(result, pd.DataFrame):
        raise ValueError(
            "The outcome of the processed text is a DataFrame, "
            "which is not supported in `process_text`."
        )

    return df.assign(**{column_name: result})



"""Backend functions for pyspark."""

from functools import wraps

try:
    from pyspark.pandas.extensions import register_dataframe_accessor

except ImportError:
    from janitor.utils import import_message

    import_message(
        submodule="spark",
        package="pyspark",
        conda_channel="conda-forge",
        pip_install=True,
    )

    # Define a dummy function to prevent NameError when pyspark is not installed
    def register_dataframe_accessor(name):
        def decorator(accessor):
            return accessor

        return decorator


def register_dataframe_method(method):
    """Register a function as a method attached to the Pyspark DataFrame.

    !!! note

        Modified based on pandas_flavor.register.

    <!--
    # noqa: DAR101 method
    # noqa: DAR201
    -->
    """

    def inner(*args, **kwargs):
        class AccessorMethod:
            def __init__(self, pyspark_obj):
                self._obj = pyspark_obj

            @wraps(method)
            def __call__(self, *args, **kwargs):
                return method(self._obj, *args, **kwargs)

        register_dataframe_accessor(method.__name__)(AccessorMethod)

        return method

    return inner()



"""Implementation of the `rle_id` function."""

import copy
from typing import Hashable, Iterable, Union

import pandas as pd
import pandas_flavor as pf
from pandas.api.types import is_list_like
from pandas.core.groupby.generic import DataFrameGroupBy

from janitor.utils import check_column


@pf.register_dataframe_groupby_method
@pf.register_dataframe_method
def rle_id(
    df: Union[pd.DataFrame, DataFrameGroupBy],
    column_names: Union[Hashable, Iterable[Hashable]],
    new_column_name: Hashable = "rle_id",
) -> Union[pd.DataFrame, DataFrameGroupBy]:
    """Generate run-length encoding IDs for consecutive identical values.

    This function assigns a unique ID to each consecutive run of identical
    values. When the value changes, the ID increments. This is useful for
    grouping consecutive runs separately, even if they have the same value.

    This method does not mutate the original DataFrame.

    Examples:
        Single column example:

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "grp": ["A", "A", "B", "B", "A", "A"],
        ...         "value": [1, 2, 3, 4, 5, 6],
        ...     }
        ... )
        >>> df
          grp  value
        0   A      1
        1   A      2
        2   B      3
        3   B      4
        4   A      5
        5   A      6
        >>> df.rle_id("grp")
          grp  value  rle_id
        0   A      1       1
        1   A      2       1
        2   B      3       2
        3   B      4       2
        4   A      5       3
        5   A      6       3

        Multiple columns example:

        >>> df = pd.DataFrame(
        ...     {
        ...         "a": [1, 1, 2, 2, 1],
        ...         "b": ["x", "x", "x", "y", "x"],
        ...         "value": [10, 20, 30, 40, 50],
        ...     }
        ... )
        >>> df
           a  b  value
        0  1  x     10
        1  1  x     20
        2  2  x     30
        3  2  y     40
        4  1  x     50
        >>> df.rle_id(["a", "b"])
           a  b  value  rle_id
        0  1  x     10       1
        1  1  x     20       1
        2  2  x     30       2
        3  2  y     40       3
        4  1  x     50       4

        Using result for grouped aggregation:

        >>> df = pd.DataFrame(
        ...     {
        ...         "grp": ["A", "A", "B", "B", "A"],
        ...         "value": [1, 2, 3, 4, 5],
        ...     }
        ... )
        >>> df.rle_id("grp").groupby(["grp", "rle_id"])["value"].sum()
        grp  rle_id
        A    1          3
             3          5
        B    2          7
        Name: value, dtype: int64

    Args:
        df: A pandas DataFrame or GroupBy object.
        column_names: A column name or an iterable of column names to
            use for computing the run-length encoding IDs.
        new_column_name: The name of the new column containing the
            run-length encoding IDs. Defaults to `"rle_id"`.

    Raises:
        ValueError: If any of the specified columns do not exist in the
            DataFrame.
        ValueError: If `new_column_name` already exists in the DataFrame.

    Returns:
        A pandas DataFrame or GroupBy object with the new column.
    """
    grp = None
    if isinstance(df, DataFrameGroupBy):
        grp = copy.copy(df)
        group_keys = grp.keys
        df = grp.obj
    else:
        group_keys = None

    if not is_list_like(column_names):
        column_names = [column_names]
    elif isinstance(column_names, tuple) and column_names in df.columns:
        column_names = [column_names]
    else:
        column_names = list(column_names)

    check_column(df, column_names)
    check_column(df, [new_column_name], present=False)

    def _values_changed(col):
        prev = col.shift()
        both_notna = col.notna() & prev.notna()
        return (both_notna & (col != prev)) | (col.isna() != prev.isna())

    def _compute_rle(frame):
        if len(column_names) == 1:
            changed = _values_changed(frame[column_names[0]])
        else:
            changed = pd.Series(False, index=frame.index)
            for col in column_names:
                changed = changed | _values_changed(frame[col])
        if len(changed) > 0:
            changed.iloc[0] = True
        return changed.cumsum()

    if group_keys is not None:
        rle_ids = df.groupby(group_keys, group_keys=False).apply(
            _compute_rle, include_groups=False
        )
        if isinstance(rle_ids, pd.DataFrame):
            rle_ids = rle_ids.stack().droplevel(0)
    else:
        rle_ids = _compute_rle(df)

    result = df.assign(**{new_column_name: rle_ids})
    if grp is not None:
        grp.obj = result
        return grp
    return result



"""Compare DataFrame column types before binding."""

from __future__ import annotations

import inspect
import json
import warnings
from collections.abc import Mapping, Sequence
from types import FrameType
from typing import Any, Iterable

import pandas as pd

_VALID_RETURN_VALUES = {"all", "match", "mismatch"}
_VALID_BIND_METHODS = {"bind_rows", "rbind"}


def describe_class(obj: Any, strict_description: bool = True) -> str:
    """Describe the class of an object or pandas series.

    This function is used by :func:`compare_df_cols` to summarize column types.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> janitor.describe_class(pd.Series([1, 2, 3]))
        'int64'
        >>> janitor.describe_class(pd.Categorical(["a", "b", "a"]))
        'category(levels=["a", "b"])'
        >>> janitor.describe_class(
        ...     pd.Categorical(["a", "b", "a"]), strict_description=False
        ... )
        'category'

    Args:
        obj: Object to describe.
        strict_description: Whether to include categorical levels in the
            description. Defaults to True.

    Returns:
        A string describing the object's class.
    """
    if isinstance(obj, pd.Categorical):
        return _describe_categorical(obj, strict_description)

    if isinstance(obj, pd.Series):
        if isinstance(obj.dtype, pd.CategoricalDtype):
            return _describe_categorical(obj, strict_description)
        return str(obj.dtype)

    dtype = getattr(obj, "dtype", None)
    if isinstance(dtype, pd.CategoricalDtype):
        return _describe_categorical(dtype, strict_description)
    if dtype is not None:
        return str(dtype)

    return type(obj).__name__


def compare_df_cols(
    *dfs: pd.DataFrame | Sequence[pd.DataFrame] | Mapping[str, pd.DataFrame],
    return_: str = "all",
    bind_method: str = "bind_rows",
    strict_description: bool = False,
    **named_dfs: pd.DataFrame | Sequence[pd.DataFrame] | Mapping[str, pd.DataFrame],
) -> pd.DataFrame:
    """Compare column types across DataFrames.

    Provide multiple DataFrames (or lists/dicts of DataFrames) to compare the
    column types prior to binding. Named arguments become output column names.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df1 = pd.DataFrame(
        ...     {
        ...         "A": [1, 2],
        ...         "B": pd.Series(["x", "y"], dtype="object"),
        ...     }
        ... )
        >>> df2 = pd.DataFrame(
        ...     {
        ...         "A": [3.0, 4.0],
        ...         "B": pd.Series(["z", "w"], dtype="object"),
        ...         "C": [True, False],
        ...     }
        ... )
        >>> df3 = pd.DataFrame(
        ...     {
        ...         "A": pd.Series(["a", "b"], dtype="object"),
        ...         "B": pd.Series(["c", "d"], dtype="object"),
        ...     }
        ... )
        >>> janitor.compare_df_cols(df1, df2, df3)
          column_name    df1     df2     df3
        0           A   int64  float64  object
        1           B  object  object  object
        2           C     NaN     bool     NaN
        >>> janitor.compare_df_cols(train=df1, test=df2)
          column_name  train    test
        0           A  int64  float64
        1           B  object  object
        2           C    NaN     bool
        >>> janitor.compare_df_cols(train=df1, test=df2, return_="mismatch")
          column_name  train    test
        0           A  int64  float64

    Args:
        *dfs: DataFrames or lists/dicts of DataFrames to compare.
        return_: Whether to return "all", only "match"ing columns, or only
            "mismatch"ing columns. Defaults to "all".
        bind_method: "bind_rows" treats missing columns as matching; "rbind"
            treats missing columns as mismatches. Defaults to "bind_rows".
        strict_description: Whether to include categorical levels in type
            descriptions. Defaults to False.
        **named_dfs: Named DataFrames or lists/dicts of DataFrames to compare.

    Raises:
        TypeError: If inputs are not DataFrames or lists/dicts of DataFrames.
        ValueError: If an invalid option is provided or no DataFrames are given.

    Returns:
        A DataFrame with a "column_name" column and one column per input
        DataFrame containing the dtype description.
    """
    if "return" in named_dfs:
        return_ = named_dfs.pop("return")

    if return_ not in _VALID_RETURN_VALUES:
        raise ValueError(
            "return_ must be one of 'all', 'match', or 'mismatch'. "
            f"Received '{return_}'."
        )
    if bind_method not in _VALID_BIND_METHODS:
        raise ValueError(
            f"bind_method must be 'bind_rows' or 'rbind'. Received '{bind_method}'."
        )

    if not dfs and not named_dfs:
        raise ValueError("At least one DataFrame is required.")

    inferred_names = _infer_arg_names(dfs)
    entries: list[tuple[str, pd.DataFrame]] = []
    for idx, (arg, inferred_name) in enumerate(zip(dfs, inferred_names, strict=True)):
        default_name = inferred_name or f"df{idx + 1}"
        entries.extend(_expand_input(arg, default_name))
    for name, arg in named_dfs.items():
        entries.extend(_expand_input(arg, name))

    if not entries:
        raise ValueError("At least one DataFrame is required.")

    for name, _ in entries:
        if name == "column_name":
            raise ValueError("None of the input names may be 'column_name'.")

    frames: list[pd.DataFrame] = []
    dataframes: list[pd.DataFrame] = []
    for name, df in entries:
        frames.append(_describe_dataframe(df, name, strict_description))
        dataframes.append(df)

    indexed = [f.set_index("column_name") for f in frames]
    result = pd.concat(indexed, axis=1, sort=False).reset_index()

    if result["column_name"].is_unique:
        column_order = _column_order(dataframes)
        if column_order:
            result = result.set_index("column_name").reindex(column_order).reset_index()

    if return_ == "all" or result.shape[1] <= 2:
        if return_ != "all":
            warnings.warn(
                "Only one DataFrame provided, so all column descriptions are returned.",
                stacklevel=2,
            )
        return result

    matches = _match_rows(result, bind_method)
    if return_ == "match":
        return result[matches].reset_index(drop=True)
    return result[~matches].reset_index(drop=True)


def compare_df_cols_same(
    *dfs: pd.DataFrame | Sequence[pd.DataFrame] | Mapping[str, pd.DataFrame],
    bind_method: str = "bind_rows",
    verbose: bool = True,
    **named_dfs: pd.DataFrame | Sequence[pd.DataFrame] | Mapping[str, pd.DataFrame],
) -> bool:
    """Check if DataFrames can be safely row-bound.

    This function returns True if there are no mismatching column types.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df1 = pd.DataFrame({"A": [1, 2]})
        >>> df2 = pd.DataFrame({"A": [3.0, 4.0]})
        >>> janitor.compare_df_cols_same(df1, df2, verbose=False)
        False

    Args:
        *dfs: DataFrames or lists/dicts of DataFrames to compare.
        bind_method: "bind_rows" treats missing columns as matching; "rbind"
            treats missing columns as mismatches. Defaults to "bind_rows".
        verbose: Whether to print mismatching columns. Defaults to True.
        **named_dfs: Named DataFrames or lists/dicts of DataFrames to compare.

    Returns:
        True if binding is safe, otherwise False.
    """
    mismatches = compare_df_cols(
        *dfs,
        bind_method=bind_method,
        return_="mismatch",
        **named_dfs,
    )
    if not mismatches.empty and verbose:
        print(mismatches)
    return mismatches.empty


def _describe_categorical(
    obj: pd.Categorical | pd.Series | pd.CategoricalDtype,
    strict_description: bool,
) -> str:
    if not strict_description:
        return "category"

    if isinstance(obj, pd.Series):
        categories = obj.cat.categories
        ordered = obj.cat.ordered
    elif isinstance(obj, pd.Categorical):
        categories = obj.categories
        ordered = obj.ordered
    else:
        categories = obj.categories
        ordered = obj.ordered

    levels_text = ", ".join(_format_level(level) for level in categories)
    category_text = f"category(levels=[{levels_text}])"
    if ordered:
        return f"ordered, {category_text}"
    return category_text


def _format_level(level: Any) -> str:
    try:
        return json.dumps(level)
    except TypeError:
        return repr(level)


def _infer_arg_names(args: Iterable[Any]) -> list[str | None]:
    frame = inspect.currentframe()
    if frame is None or frame.f_back is None:
        return [None for _ in args]
    try:
        caller = frame.f_back
        names: list[str | None] = []
        for arg in args:
            names.append(_find_name_in_frame(arg, caller))
        return names
    finally:
        del frame


def _find_name_in_frame(value: Any, frame: FrameType) -> str | None:
    for scope in (frame.f_locals, frame.f_globals):
        for name, candidate in scope.items():
            if candidate is value:
                return name
    return None


def _expand_input(
    arg: pd.DataFrame | Sequence[pd.DataFrame] | Mapping[str, pd.DataFrame],
    base_name: str | None,
) -> list[tuple[str, pd.DataFrame]]:
    if isinstance(arg, pd.DataFrame):
        name = _validate_name(base_name)
        return [(name, arg)]

    if isinstance(arg, Mapping):
        entries = []
        for idx, (name, df) in enumerate(arg.items(), start=1):
            if name is None or name == "":
                name = f"{_validate_name(base_name)}_{idx}"
            name = _validate_name(name)
            if not isinstance(df, pd.DataFrame):
                raise TypeError(
                    "All mapping values must be pandas DataFrames. "
                    f"Received {type(df).__name__}."
                )
            entries.append((name, df))
        return entries

    if isinstance(arg, Sequence):
        entries = []
        for idx, df in enumerate(arg, start=1):
            if not isinstance(df, pd.DataFrame):
                raise TypeError(
                    "All sequence items must be pandas DataFrames. "
                    f"Received {type(df).__name__}."
                )
            name = _validate_name(base_name)
            entries.append((f"{name}_{idx}", df))
        return entries

    raise TypeError(
        "Inputs must be pandas DataFrames or sequences/mappings of DataFrames. "
        f"Received {type(arg).__name__}."
    )


def _validate_name(name: str | None) -> str:
    if name is None or name == "":
        return "df"
    if not isinstance(name, str):
        raise TypeError("DataFrame names must be strings.")
    return name


def _describe_dataframe(
    df: pd.DataFrame, name: str, strict_description: bool
) -> pd.DataFrame:
    if df.shape[1] == 0:
        warnings.warn(
            f"{name} has zero columns and will not appear in output.",
            stacklevel=2,
        )
        return pd.DataFrame({"column_name": []})

    column_names = []
    descriptions = []
    for column_name, series in df.items():
        column_names.append(column_name)
        descriptions.append(
            describe_class(series, strict_description=strict_description)
        )

    return pd.DataFrame(
        {
            "column_name": column_names,
            name: descriptions,
        }
    )


def _column_order(dataframes: Iterable[pd.DataFrame]) -> list[Any]:
    order: list[Any] = []
    for df in dataframes:
        for column in df.columns:
            if column not in order:
                order.append(column)
    return order


def _match_rows(result: pd.DataFrame, bind_method: str) -> pd.Series:
    data = result.iloc[:, 1:]
    if bind_method == "rbind":
        return data.apply(_row_matches_rbind, axis=1)
    return data.apply(_row_matches_bind_rows, axis=1)


def _row_matches_rbind(row: pd.Series) -> bool:
    values = row.to_list()
    first = values[0]
    if pd.isna(first):
        return False
    return all(value == first for value in values[1:])


def _row_matches_bind_rows(row: pd.Series) -> bool:
    values = [value for value in row.to_list() if not pd.isna(value)]
    if len(values) <= 1:
        return True
    first = values[0]
    return all(value == first for value in values[1:])



"""Implementation of remove_columns."""

from typing import Hashable, Iterable, Union

import pandas as pd
import pandas_flavor as pf

from janitor.utils import deprecated_alias, refactored_function


@pf.register_dataframe_method
@refactored_function(
    message=(
        "This function will be deprecated in a 1.x release. "
        "Please use `pd.DataFrame.drop` instead."
    )
)
@deprecated_alias(columns="column_names")
def remove_columns(
    df: pd.DataFrame,
    column_names: Union[str, Iterable[str], Hashable],
) -> pd.DataFrame:
    """Remove the set of columns specified in `column_names`.

    This method does not mutate the original DataFrame.

    Intended to be the method-chaining alternative to `del df[col]`.

    !!!note

        This function will be deprecated in a 1.x release.
        Kindly use `pd.DataFrame.drop` instead.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": [2, 4, 6], "b": [1, 3, 5], "c": [7, 8, 9]})
        >>> df
           a  b  c
        0  2  1  7
        1  4  3  8
        2  6  5  9
        >>> df.remove_columns(column_names=["a", "c"])
           b
        0  1
        1  3
        2  5

    Args:
        df: A pandas DataFrame.
        column_names: The columns to remove.

    Returns:
        A pandas DataFrame.
    """

    return df.drop(columns=column_names)



"""Finance-specific data cleaning functions."""

import json
from datetime import date
from functools import lru_cache
from typing import Optional

import pandas as pd
import pandas_flavor as pf
import requests

from janitor.errors import JanitorError

from .utils import check, deprecated_alias, is_connected

currency_set = {
    "AUD",
    "BGN",
    "BRL",
    "CAD",
    "CHF",
    "CNY",
    "CZK",
    "DKK",
    "EUR",
    "GBP",
    "HKD",
    "HRK",
    "HUF",
    "IDR",
    "ILS",
    "INR",
    "ISK",
    "JPY",
    "KRW",
    "MXN",
    "MYR",
    "NOK",
    "NZD",
    "PHP",
    "PLN",
    "RON",
    "RUB",
    "SEK",
    "SGD",
    "THB",
    "TRY",
    "USD",
    "ZAR",
}

# Dictionary of recognized World Bank countries and their abbreviations
wb_country_dict = {
    "Aruba": "ABW",
    "Afghanistan": "AFG",
    "Angola": "AGO",
    "Albania": "ALB",
    "Andorra": "AND",
    "Arab World": "ARB",
    "United Arab Emirates": "ARE",
    "Argentina": "ARG",
    "Armenia": "ARM",
    "American Samoa": "ASM",
    "Antigua and Barbuda": "ATG",
    "Australia": "AUS",
    "Austria": "AUT",
    "Azerbaijan": "AZE",
    "Burundi": "BDI",
    "Belgium": "BEL",
    "Benin": "BEN",
    "Burkina Faso": "BFA",
    "Bangladesh": "BGD",
    "Bulgaria": "BGR",
    "Bahrain": "BHR",
    "Bahamas, The": "BHS",
    "Bosnia and Herzegovina": "BIH",
    "Belarus": "BLR",
    "Belize": "BLZ",
    "Bermuda": "BMU",
    "Bolivia": "BOL",
    "Brazil": "BRA",
    "Barbados": "BRB",
    "Brunei Darussalam": "BRN",
    "Bhutan": "BTN",
    "Botswana": "BWA",
    "Central African Republic": "CAF",
    "Canada": "CAN",
    "Central Europe and the Baltics": "CEB",
    "Switzerland": "CHE",
    "Channel Islands": "CHI",
    "Chile": "CHL",
    "China": "CHN",
    "Cote d'Ivoire": "CIV",
    "Cameroon": "CMR",
    "Congo, Dem. Rep.": "COD",
    "Congo, Rep.": "COG",
    "Colombia": "COL",
    "Comoros": "COM",
    "Cabo Verde": "CPV",
    "Costa Rica": "CRI",
    "Caribbean small states": "CSS",
    "Cuba": "CUB",
    "Curacao": "CUW",
    "Cayman Islands": "CYM",
    "Cyprus": "CYP",
    "Czech Republic": "CZE",
    "Germany": "DEU",
    "Djibouti": "DJI",
    "Dominica": "DMA",
    "Denmark": "DNK",
    "Dominican Republic": "DOM",
    "Algeria": "DZA",
    "East Asia & Pacific (excluding high income)": "EAP",
    "Early-demographic dividend": "EAR",
    "East Asia & Pacific": "EAS",
    "Europe & Central Asia (excluding high income)": "ECA",
    "Europe & Central Asia": "ECS",
    "Ecuador": "ECU",
    "Egypt, Arab Rep.": "EGY",
    "Euro area": "EMU",
    "Eritrea": "ERI",
    "Spain": "ESP",
    "Estonia": "EST",
    "Ethiopia": "ETH",
    "European Union": "EUU",
    "Fragile and conflict affected situations": "FCS",
    "Finland": "FIN",
    "Fiji": "FJI",
    "France": "FRA",
    "Faroe Islands": "FRO",
    "Micronesia, Fed. Sts.": "FSM",
    "Gabon": "GAB",
    "United Kingdom": "GBR",
    "Georgia": "GEO",
    "Ghana": "GHA",
    "Gibraltar": "GIB",
    "Guinea": "GIN",
    "Gambia, The": "GMB",
    "Guinea-Bissau": "GNB",
    "Equatorial Guinea": "GNQ",
    "Greece": "GRC",
    "Grenada": "GRD",
    "Greenland": "GRL",
    "Guatemala": "GTM",
    "Guam": "GUM",
    "Guyana": "GUY",
    "High income": "HIC",
    "Hong Kong SAR, China": "HKG",
    "Honduras": "HND",
    "Heavily indebted poor countries (HIPC)": "HPC",
    "Croatia": "HRV",
    "Haiti": "HTI",
    "Hungary": "HUN",
    "IBRD only": "IBD",
    "IDA & IBRD total": "IBT",
    "IDA total": "IDA",
    "IDA blend": "IDB",
    "Indonesia": "IDN",
    "IDA only": "IDX",
    "Isle of Man": "IMN",
    "India": "IND",
    "Not classified": "INX",
    "Ireland": "IRL",
    "Iran, Islamic Rep.": "IRN",
    "Iraq": "IRQ",
    "Iceland": "ISL",
    "Israel": "ISR",
    "Italy": "ITA",
    "Jamaica": "JAM",
    "Jordan": "JOR",
    "Japan": "JPN",
    "Kazakhstan": "KAZ",
    "Kenya": "KEN",
    "Kyrgyz Republic": "KGZ",
    "Cambodia": "KHM",
    "Kiribati": "KIR",
    "St. Kitts and Nevis": "KNA",
    "Korea, Rep.": "KOR",
    "Kuwait": "KWT",
    "Latin America & Caribbean (excluding high income)": "LAC",
    "Lao PDR": "LAO",
    "Lebanon": "LBN",
    "Liberia": "LBR",
    "Libya": "LBY",
    "St. Lucia": "LCA",
    "Latin America & Caribbean": "LCN",
    "Least developed countries: UN classification": "LDC",
    "Low income": "LIC",
    "Liechtenstein": "LIE",
    "Sri Lanka": "LKA",
    "Lower middle income": "LMC",
    "Low & middle income": "LMY",
    "Lesotho": "LSO",
    "Late-demographic dividend": "LTE",
    "Lithuania": "LTU",
    "Luxembourg": "LUX",
    "Latvia": "LVA",
    "Macao SAR, China": "MAC",
    "St. Martin (French part)": "MAF",
    "Morocco": "MAR",
    "Monaco": "MCO",
    "Moldova": "MDA",
    "Madagascar": "MDG",
    "Maldives": "MDV",
    "Middle East & North Africa": "MEA",
    "Mexico": "MEX",
    "Marshall Islands": "MHL",
    "Middle income": "MIC",
    "North Macedonia": "MKD",
    "Mali": "MLI",
    "Malta": "MLT",
    "Myanmar": "MMR",
    "Middle East & North Africa (excluding high income)": "MNA",
    "Montenegro": "MNE",
    "Mongolia": "MNG",
    "Northern Mariana Islands": "MNP",
    "Mozambique": "MOZ",
    "Mauritania": "MRT",
    "Mauritius": "MUS",
    "Malawi": "MWI",
    "Malaysia": "MYS",
    "North America": "NAC",
    "Namibia": "NAM",
    "New Caledonia": "NCL",
    "Niger": "NER",
    "Nigeria": "NGA",
    "Nicaragua": "NIC",
    "Netherlands": "NLD",
    "Norway": "NOR",
    "Nepal": "NPL",
    "Nauru": "NRU",
    "New Zealand": "NZL",
    "OECD members": "OED",
    "Oman": "OMN",
    "Other small states": "OSS",
    "Pakistan": "PAK",
    "Panama": "PAN",
    "Peru": "PER",
    "Philippines": "PHL",
    "Palau": "PLW",
    "Papua New Guinea": "PNG",
    "Poland": "POL",
    "Pre-demographic dividend": "PRE",
    "Puerto Rico": "PRI",
    "Korea, Dem. People's Rep.": "PRK",
    "Portugal": "PRT",
    "Paraguay": "PRY",
    "West Bank and Gaza": "PSE",
    "Pacific island small states": "PSS",
    "Post-demographic dividend": "PST",
    "French Polynesia": "PYF",
    "Qatar": "QAT",
    "Romania": "ROU",
    "Russian Federation": "RUS",
    "Rwanda": "RWA",
    "South Asia": "SAS",
    "Saudi Arabia": "SAU",
    "Sudan": "SDN",
    "Senegal": "SEN",
    "Singapore": "SGP",
    "Solomon Islands": "SLB",
    "Sierra Leone": "SLE",
    "El Salvador": "SLV",
    "San Marino": "SMR",
    "Somalia": "SOM",
    "Serbia": "SRB",
    "Sub-Saharan Africa (excluding high income)": "SSA",
    "South Sudan": "SSD",
    "Sub-Saharan Africa": "SSF",
    "Small states": "SST",
    "Sao Tome and Principe": "STP",
    "Suriname": "SUR",
    "Slovak Republic": "SVK",
    "Slovenia": "SVN",
    "Sweden": "SWE",
    "Eswatini": "SWZ",
    "Sint Maarten (Dutch part)": "SXM",
    "Seychelles": "SYC",
    "Syrian Arab Republic": "SYR",
    "Turks and Caicos Islands": "TCA",
    "Chad": "TCD",
    "East Asia & Pacific (IDA & IBRD countries)": "TEA",
    "Europe & Central Asia (IDA & IBRD countries)": "TEC",
    "Togo": "TGO",
    "Thailand": "THA",
    "Tajikistan": "TJK",
    "Turkmenistan": "TKM",
    "Latin America & the Caribbean (IDA & IBRD countries)": "TLA",
    "Timor-Leste": "TLS",
    "Middle East & North Africa (IDA & IBRD countries)": "TMN",
    "Tonga": "TON",
    "South Asia (IDA & IBRD)": "TSA",
    "Sub-Saharan Africa (IDA & IBRD countries)": "TSS",
    "Trinidad and Tobago": "TTO",
    "Tunisia": "TUN",
    "Turkey": "TUR",
    "Tuvalu": "TUV",
    "Tanzania": "TZA",
    "Uganda": "UGA",
    "Ukraine": "UKR",
    "Upper middle income": "UMC",
    "Uruguay": "URY",
    "United States": "USA",
    "Uzbekistan": "UZB",
    "St. Vincent and the Grenadines": "VCT",
    "Venezuela, RB": "VEN",
    "British Virgin Islands": "VGB",
    "Virgin Islands (U.S.)": "VIR",
    "Vietnam": "VNM",
    "Vanuatu": "VUT",
    "World": "WLD",
    "Samoa": "WSM",
    "Kosovo": "XKX",
    "Yemen, Rep.": "YEM",
    "South Africa": "ZAF",
    "Zambia": "ZMB",
    "Zimbabwe": "ZWE",
}


def _check_currency(currency: str):
    """Check that currency is in supported set."""
    if currency not in currency_set:
        raise ValueError(
            f"currency {currency} not in supported currency set, {currency_set}"
        )


def _check_wb_country(country: str):
    """Check that world bank country is in supported set."""
    if (country not in wb_country_dict.keys()) & (
        country not in wb_country_dict.values()  # noqa: PD011
    ):
        raise ValueError(
            f"country {country} not in supported World Bank country dict, "
            f"{wb_country_dict}"
        )


def _check_wb_years(year: int):
    """Check that year is in world bank dataset years."""
    if year < 1960:
        raise ValueError("year value must be 1960 or later")


# @lru_cache(maxsize=32)
# def _convert_currency(
#     api_key: str,
#     from_currency: str = None,
#     to_currency: str = None,
#     historical_date: Optional[date] = None,
# ) -> float:
#     """
#     Currency conversion for Pandas DataFrame column.

#     Helper function for `convert_currency` method.

#     The API used is https://exchangeratesapi.io/.
#     """

#     url = "http://api.exchangeratesapi.io"

#     if historical_date:
#         check("historical_date", historical_date, [datetime, date])
#         if isinstance(historical_date, datetime):
#             if historical_date < datetime(1999, 1, 4):
#                 raise ValueError(
#                     "historical_date:datetime must be later than 1999-01-04!"
#                 )
#             string_date = str(historical_date)[:10]
#         else:
#             if historical_date < date(1999, 1, 4):
#                 raise ValueError(
#                     "historical_date:date must be later than 1999-01-04!"
#                 )
#             string_date = str(historical_date)
#         url = url + "/%s" % string_date
#     else:
#         url = url + "/latest"

#     _check_currency(from_currency)
#     _check_currency(to_currency)

#     payload = {
#         # "base": from_currency,
#         "symbols": to_currency,
#         "access_key": api_key,
#     }

#     result = requests.get(url, params=payload)

#     if result.status_code != 200:
#         raise ConnectionError(
#             "Exchange Rate API failed to receive a 200 "
#             "response from the server. "
#             "Please try again later."
#         )

#     currency_dict = json.loads(result.text)
#     rate = currency_dict["rates"][to_currency]
#     return rate


@pf.register_dataframe_method
@deprecated_alias(colname="column_name")
def convert_currency(
    df: pd.DataFrame,
    api_key: str,
    column_name: str = None,
    from_currency: str = None,
    to_currency: str = None,
    historical_date: date = None,
    make_new_column: bool = False,
) -> pd.DataFrame:
    """Deprecated function.

    <!--
    # noqa: DAR101
    # noqa: DAR401
    -->
    """
    raise JanitorError(
        "The `convert_currency` function has been temporarily disabled due to "
        "exchangeratesapi.io disallowing free pinging of its API. "
        "(Our tests started to fail due to this issue.) "
        "There is no easy way around this problem "
        "except to find a new API to call on."
        "Please comment on issue #829 "
        "(https://github.com/pyjanitor-devs/pyjanitor/issues/829) "
        "if you know of an alternative API that we can call on, "
        "otherwise the function will be removed in pyjanitor's 1.0 release."
    )


# @pf.register_dataframe_method
# @deprecated_alias(colname="column_name")
# def convert_currency(
#     df: pd.DataFrame,
#     api_key: str,
#     column_name: str = None,
#     from_currency: str = None,
#     to_currency: str = None,
#     historical_date: date = None,
#     make_new_column: bool = False,
# ) -> pd.DataFrame:
#     """
#     Converts a column from one currency to another, with an option to
#     convert based on historical exchange values.

#     On April 10 2021,
#     we discovered that there was no more free API available.
#     Thus, an API key is required to perform currency conversion.
#     API keys should be set as an environment variable,
#     for example, `EXCHANGE_RATE_API_KEY``,
#     and then passed into the function
#     by calling on `os.getenv("EXCHANGE_RATE_APIKEY")``.

#     :param df: A pandas dataframe.
#     :param api_key: exchangeratesapi.io API key.
#     :param column_name: Name of the new column. Should be a string, in order
#         for the column name to be compatible with the Feather binary
#         format (this is a useful thing to have).
#     :param from_currency: The base currency to convert from.
#         May be any of: currency_set = {"AUD", "BGN", "BRL", "CAD", "CHF",
#         "CNY", "CZK", "DKK", "EUR", "GBP", "HKD", "HRK", "HUF", "IDR",
#         "ILS", "INR", "ISK", "JPY", "KRW", "MXN", "MYR", "NOK", "NZD",
#         "PHP", "PLN", "RON", "RUB", "SEK", "SGD", "THB", "TRY", "USD",
#         "ZAR"}
#     :param to_currency: The target currency to convert to.
#         May be any of: currency_set = {"AUD", "BGN", "BRL", "CAD", "CHF",
#         "CNY", "CZK", "DKK", "EUR", "GBP", "HKD", "HRK", "HUF", "IDR",
#         "ILS", "INR", "ISK", "JPY", "KRW", "MXN", "MYR", "NOK", "NZD",
#         "PHP", "PLN", "RON", "RUB", "SEK", "SGD", "THB", "TRY", "USD",
#         "ZAR"}
#     :param historical_date: If supplied,
#         get exchange rate on a certain date.
#         If not supplied, get the latest exchange rate.
#         The exchange rates go back to Jan. 4, 1999.
#     :param make_new_column: Generates new column
#         for converted currency if True,
#         otherwise, converts currency in place.
#     :returns: The dataframe with converted currency column.

#     .. code-block:: python

#         import pandas as pd
#         import janitor
#         from datetime import date

#         data_dict = {
#             "a": [1.23452345, 2.456234, 3.2346125] * 3,
#             "Bell__Chart": [1/3, 2/7, 3/2] * 3,
#             "decorated-elephant": [1/234, 2/13, 3/167] * 3,
#             "animals": ["rabbit", "leopard", "lion"] * 3,
#             "cities": ["Cambridge", "Shanghai", "Basel"] * 3,
#         }

#         example_dataframe = pd.DataFrame(data_dict)

#     Example: Converting a column from one currency to another
#     using rates from 01/01/2018.

#     .. code-block:: python

#         example_dataframe.convert_currency('a', from_currency='USD',
#         to_currency='EUR', historical_date=date(2018,1,1))

#     Output:

#     .. code-block:: python

#                     a  Bell__Chart  decorated-elephant  animals     cities
#         0  1.029370     0.333333            0.004274   rabbit  Cambridge
#         1  2.048056     0.285714            0.153846  leopard   Shanghai
#         2  2.697084     1.500000            0.017964     lion      Basel
#         3  1.029370     0.333333            0.004274   rabbit  Cambridge
#         4  2.048056     0.285714            0.153846  leopard   Shanghai
#         5  2.697084     1.500000            0.017964     lion      Basel
#         6  1.029370     0.333333            0.004274   rabbit  Cambridge
#         7  2.048056     0.285714            0.153846  leopard   Shanghai
#         8  2.697084     1.500000            0.017964     lion      Basel
#     """

#     rate = _convert_currency(
#         api_key, from_currency, to_currency, historical_date
#     )

#     if make_new_column:
#         # new_column_name = column_name + "_" + to_currency
#         column_name = column_name + "_" + to_currency

#     df = df.assign(column_name=df[column_name] * rate)

#     return df


@lru_cache(maxsize=32)
def _inflate_currency(
    country: str = None, currency_year: int = None, to_year: int = None
) -> float:
    """
    Currency inflation for Pandas DataFrame column.
    Helper function for `inflate_currency` method.
    The API used is the World Bank Indicator API:
    https://datahelpdesk.worldbank.org/knowledgebase/articles/889392-about-the-indicators-api-documentation
    """

    # Check all inputs are correct data type
    check("country", country, [str])
    check("currency_year", currency_year, [int])
    check("to_year", to_year, [int])

    # Get WB country abbreviation
    _check_wb_country(country)
    if country in wb_country_dict.keys():
        country = wb_country_dict[country]
    else:
        # `country` is already a correct abbreviation; do nothing
        pass

    _check_wb_years(currency_year)
    _check_wb_years(to_year)

    url = (
        "https://api.worldbank.org/v2/country/"
        + country
        + "/indicator/FP.CPI.TOTL?date="
        + str(min(currency_year, to_year))
        + ":"
        + str(max(currency_year, to_year))
        + "&format=json"
    )

    result = requests.get(url)

    if result.status_code != 200:
        raise ConnectionError(
            "WB Indicator API failed to receive a 200 "
            "response from the server. "
            "Please try again later."
        )

    # The API returns a list of two items;
    # the second item in the list is what we want
    inflation_dict = json.loads(result.text)[1]

    # Error checking
    if inflation_dict is None:
        raise ValueError(
            "The WB Indicator API returned nothing. "
            "This likely means the currency_year and "
            "to_year are outside of the year range for "
            "which the WB has inflation data for the "
            "specified country."
        )

    # Create new dict with only the year and inflation values
    inflation_dict_ready = {
        int(inflation_dict[i]["date"]): float(inflation_dict[i]["value"])
        for i in range(len(inflation_dict))
        if inflation_dict[i]["value"] is not None
    }

    # Error catching
    if currency_year not in inflation_dict_ready.keys():
        raise ValueError(
            f"The WB Indicator API does not have inflation "
            f"data for {currency_year} for {country}."
        )
    if to_year not in inflation_dict_ready.keys():
        raise ValueError(
            f"The WB Indicator API does not have inflation "
            f"data for {to_year} for {country}."
        )

    inflator = inflation_dict_ready[to_year] / inflation_dict_ready[currency_year]
    return inflator


@pf.register_dataframe_method
def inflate_currency(
    df: pd.DataFrame,
    column_name: str = None,
    country: str = None,
    currency_year: int = None,
    to_year: int = None,
    make_new_column: bool = False,
) -> pd.DataFrame:
    """
    Inflates a column of monetary values from one year to another, based on
    the currency's country.

    The provided country can be any economy name or code from the World Bank
    [list of economies](https://databank.worldbank.org/data/download/site-content/CLASS.xls).

    **Note**: This method mutates the original DataFrame.

    Examples:

        >>> import pandas as pd
        >>> import janitor.finance
        >>> df = pd.DataFrame({"profit": [100.10, 200.20, 300.30, 400.40, 500.50]})
        >>> df
           profit
        0   100.1
        1   200.2
        2   300.3
        3   400.4
        4   500.5
        >>> df.inflate_currency(
        ...     column_name="profit",
        ...     country="USA",
        ...     currency_year=2015,
        ...     to_year=2018,
        ...     make_new_column=True,
        ... )
           profit  profit_2018
        0   100.1   106.050596
        1   200.2   212.101191
        2   300.3   318.151787
        3   400.4   424.202382
        4   500.5   530.252978

    Args:
        df: A pandas DataFrame.
        column_name: Name of the column containing monetary
            values to inflate.
        country: The country associated with the currency being inflated.
            May be any economy or code from the World Bank
            [List of economies](https://databank.worldbank.org/data/download/site-content/CLASS.xls).
        currency_year: The currency year to inflate from.
            The year should be 1960 or later.
        to_year: The currency year to inflate to.
            The year should be 1960 or later.
        make_new_column: Generates new column for inflated currency if
            True, otherwise, inflates currency in place.

    Returns:
        The DataFrame with inflated currency column.
    """  # noqa: E501

    inflator = _inflate_currency(country, currency_year, to_year)

    if make_new_column:
        new_column_name = column_name + "_" + str(to_year)
        df[new_column_name] = df[column_name] * inflator

    else:
        df[column_name] = df[column_name] * inflator

    return df


def convert_stock(stock_symbol: str) -> str:
    """
    This function takes in a stock symbol as a parameter,
    queries an API for the companies full name and returns
    it

    Examples:

        ```python
        import janitor.finance

        janitor.finance.convert_stock("aapl")
        ```

    Args:
        stock_symbol: Stock ticker Symbol

    Raises:
        ConnectionError: Internet connection is not available

    Returns:
        Full company name
    """
    if is_connected("www.google.com"):
        stock_symbol = stock_symbol.upper()
        return get_symbol(stock_symbol)
    else:
        raise ConnectionError("Connection Error: Client Not Connected to Internet")


def get_symbol(symbol: str) -> Optional[str]:
    """
    This is a helper function to get a companies full
    name based on the stock symbol.

    Examples:

        ```python
        import janitor.finance

        janitor.finance.get_symbol("aapl")
        ```

    Args:
        symbol: This is our stock symbol that we use
            to query the api for the companies full name.

    Returns:
        Company full name
    """
    result = requests.get(
        "http://d.yimg.com/autoc."
        + "finance.yahoo.com/autoc?query={}&region=1&lang=en".format(symbol)
    ).json()

    for x in result["ResultSet"]["Result"]:
        if x["symbol"] == symbol:
            return x["name"]
        else:
            return None



"""Implementation of summarise."""

from __future__ import annotations

from functools import singledispatch
from typing import Any

import pandas as pd
import pandas_flavor as pf
from pandas.api.types import is_scalar
from pandas.core.common import apply_if_callable
from pandas.core.groupby.generic import DataFrameGroupBy

from janitor.functions.select import get_index_labels
from janitor.utils import refactored_function


@pf.register_dataframe_groupby_method
@pf.register_dataframe_method
@refactored_function(
    message=(
        "This function is deprecated. "
        "Please use `pd.DataFrame.agg` instead."
    )
)
def summarise(
    df: pd.DataFrame | DataFrameGroupBy,
    *args: tuple[dict | tuple],
) -> pd.DataFrame:
    """

    !!! info "New in version 0.31.0"

    !!!note

        Before reaching for `summarise`, try `pd.DataFrame.agg`.

    summarise creates a new dataframe;
    it returns one row for each combination of grouping columns.
    If there are no grouping variables,
    the output will have a single row
    summarising all observations in the input.

    The argument provided to *args* should be either
    a dictionary, a callable or a tuple; however,
    anything can be passed, as long as it fits
    within pandas' aggregation semantics.

    - **dictionary argument**:
    If the argument is a dictionary,
    the value in the `{key:value}` pairing
    should be either a string, a callable, or a tuple.

        - If the value in the dictionary
        is a string or a callable,
        the key of the dictionary
        should be an existing column name.

        The function is applied on the `df[column_name]` series.

        !!!note

            - If the value is a string,
            the string should be a pandas string function,
            e.g "sum", "mean", etc.

        - If the value of the dictionary is a tuple,
        it should be of length 2, and of the form
        `(column_name, aggfunc)`,
        where `column_name` should exist in the DataFrame,
        and `aggfunc` should be either a string or a callable.

        This option allows for custom renaming of the aggregation output,
        where the key in the dictionary can be a new column name.


    - **tuple argument**:
    If the argument is a tuple, it should be of length 2,
    and of the form
    `(column_name, aggfunc)`,
    where column_name should exist in the DataFrame,
    and `aggfunc` should be either a string or a callable.

        !!!note

            - if `aggfunc` is a string,
            the string should be a pandas string function,
            e.g "sum", "mean", etc.

        !!!note

            - `column_name` can be anything supported by the
            [`select`][janitor.functions.select.select] syntax;
            as such multiple columns can be processed here -
            they will be processed individually.


    - **callable argument**:
    If the argument is a callable, the callable is applied
    on the DataFrame or GroupBy object.
    The result from the callable should be a pandas Series
    or DataFrame.


    Aggregated columns cannot be reused in `summarise`.


    Arguments supported in `pd.DataFrame.groupby`
    can also be passed to `by` via a dictionary.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> data = {
        ...     "avg_jump": [3, 4, 1, 2, 3, 4],
        ...     "avg_run": [3, 4, 1, 3, 2, 4],
        ...     "combine_id": [100200, 100200, 101200, 101200, 102201, 103202],
        ... }
        >>> df = pd.DataFrame(data)
        >>> df
           avg_jump  avg_run  combine_id
        0         3        3      100200
        1         4        4      100200
        2         1        1      101200
        3         2        3      101200
        4         3        2      102201
        5         4        4      103202

        Aggregation on a DataFrame via a callable:
        >>> df.summarise(
        ...     lambda df: df.select("avg*").mean().rename("mean")
        ... )  # doctest: +NORMALIZE_WHITESPACE
                             mean
        avg_jump         2.833333
        avg_run          2.833333

        Aggregation on a DataFrame via a tuple:
        >>> df.summarise(("avg_*", "mean"))
              avg_jump   avg_run
        mean  2.833333  2.833333

        Aggregation on a DataFrame via a dictionary:
        >>> df.summarise({"avg_jump": "mean"})
              avg_jump
        mean  2.833333

        >>> df.summarise({"avg_run_2": ("avg_run", "mean")})
              avg_run_2
        mean   2.833333

        >>> grouped = df.groupby("combine_id")

        Aggregation on a grouped object via a callable:
        >>> grouped.summarise(lambda df: df.sum())  # doctest: +NORMALIZE_WHITESPACE
                 avg_jump  avg_run
        combine_id
        100200             7        7
        101200             3        4
        102201             3        2
        103202             4        4

        Aggregation on a grouped object via a tuple:
        >>> grouped.summarise(("avg_run", "mean"))  # doctest: +NORMALIZE_WHITESPACE
                    avg_run
        combine_id
        100200          3.5
        101200          2.0
        102201          2.0
        103202          4.0

        Aggregation on a grouped object via a dictionary:
        >>> grouped.summarise({"avg_run": "mean"})  # doctest: +NORMALIZE_WHITESPACE
                    avg_run
        combine_id
        100200          3.5
        101200          2.0
        102201          2.0
        103202          4.0
        >>> grouped.summarise(
        ...     {"avg_run_2": ("avg_run", "mean")}
        ... )  # doctest: +NORMALIZE_WHITESPACE
                    avg_run_2
        combine_id
        100200            3.5
        101200            2.0
        102201            2.0
        103202            4.0

    Args:
        df: A pandas DataFrame or DataFrameGroupBy object.
        args: Either a dictionary or a tuple.

    Raises:
        ValueError: If a tuple is passed and the length is not 2.

    Returns:
        A pandas DataFrame with aggregated columns.

    """  # noqa: E501
    if isinstance(df, DataFrameGroupBy):
        by = df
        df = df.obj
    else:
        by = None
    contents = []
    for arg in args:
        aggregate = _aggfunc(arg, df=df, by=by)
        contents.extend(aggregate)
    counts = 0
    for entry in contents:
        if isinstance(entry, pd.DataFrame):
            length = entry.columns.nlevels
        elif isinstance(entry.name, tuple):
            length = len(entry.name)
        else:
            length = 1
        counts = max(counts, length)
    contents_ = []
    for entry in contents:
        if isinstance(entry, pd.DataFrame):
            length_ = entry.columns.nlevels
            length = counts - length_
            if length:
                patch = [""] * length
                columns = [entry.columns.get_level_values(n) for n in range(length_)]
                columns.append(patch)
                names = [*entry.columns.names]
                names.extend([None] * length)
                columns = pd.MultiIndex.from_arrays(columns, names=names)
                entry.columns = columns
        elif is_scalar(entry.name):
            length = counts - 1
            if length:
                patch = [""] * length
                name = (entry.name, *patch)
                entry.name = name
        elif isinstance(entry.name, tuple):
            length = counts - len(entry.name)
            if length:
                patch = [""] * length
                name = (*entry.name, *patch)
                entry.name = name
        contents_.append(entry)
    return pd.concat(contents_, axis=1,  sort=False)


@singledispatch
def _aggfunc(arg, df, by):
    if by is None:
        val = df
    else:
        val = by
    outcome = apply_if_callable(maybe_callable=arg, obj=val)
    if isinstance(outcome, pd.Series):
        if not outcome.name:
            raise ValueError("Ensure the pandas Series object has a name")
        return [outcome]
    if isinstance(outcome, pd.DataFrame):
        return [outcome]
    raise TypeError(
        "The output from the aggregation should be a named Series or a DataFrame"
    )


@_aggfunc.register(tuple)
def _(arg, df, by):
    """Dispatch function for tuple"""
    if len(arg) != 2:
        raise ValueError("the tuple has to be a length of 2")
    column_name, aggfunc = arg
    column_names = get_index_labels(arg=[column_name], df=df, axis="columns")
    mapping = {column_name: aggfunc for column_name in column_names}
    return _aggfunc(mapping, df=df, by=by)


@_aggfunc.register(dict)
def _(arg, df, by):
    """Dispatch function for dictionary"""
    if by is None:
        val = df
    else:
        val = by

    contents = []
    for column_name, aggfunc in arg.items():
        if isinstance(aggfunc, tuple):
            if len(aggfunc) != 2:
                raise ValueError("the tuple has to be a length of 2")
            column, func = aggfunc
            column = val.agg({column: func})
            try:
                column = column.squeeze()
            except AttributeError:
                pass
            column = _convert_obj_to_named_series(
                obj=column,
                column_name=column_name,
                function=func,
            )
            if not isinstance(column, pd.Series):
                raise TypeError(
                    f"Expected a pandas Series object; instead got {type(column)}"
                )
        else:
            column = val.agg({column_name: aggfunc})
            try:
                column = column.squeeze()
            except AttributeError:
                pass
            column = _convert_obj_to_named_series(
                obj=column,
                column_name=column_name,
                function=aggfunc,
            )
        contents.append(column)
    return contents


def _convert_obj_to_named_series(obj, function: Any, column_name: Any):
    if isinstance(obj, pd.Series):
        obj.name = column_name
        return obj
    if not is_scalar(obj):
        return obj
    if isinstance(function, str):
        function_name = function
    else:
        function_name = function.__name__
    return pd.Series(data=obj, index=[function_name], name=column_name)
